#include "stdafx.h"
#include "PEDAsm.h"


PCSTR PEDisAsm::G_PREFIXES[] =
{
	"REP", "REPZ", "REPN", "REPNZ", "LOCK", 
	"CS", "DS", "ES", "SS", "FS", "GS"
};

PCSTR PEDisAsm::G_REGISTERS[] =
{
	// REGS64_BASE : 0
	"RAX" , "RCX" , "RDX"  , "RBX"  , "RSP"  , "RBP"  , "RSI"  , "RDI"	,
	// EX_GPR_BASE : 8
	"R8"  , "R9"  , "R10"  , "R11"  , "R12"  , "R13"  , "R14"  , "R15"  ,
	// REGS32_BASE : 16
	"EAX" , "ECX" , "EDX"  , "EBX"  , "ESP"  , "EBP"  , "ESI"  , "EDI"  , 
	"R8D" , "R9D" , "R10D" , "R11D" , "R12D" , "R13D" , "R14D" , "R15D" ,
	// REGS16_BASE : 32
	"AX"  , "CX"  , "DX"   , "BX"   , "SP"   , "BP"   , "SI"   , "DI"   , 
	"R8W" , "R9W" , "R10W" , "R11W" , "R12W" , "R13W" , "R14W" , "R15W" ,
	// REGS8_BASE : 48
	"AL"  , "CL"  , "DL"   , "BL"   , "AH"   , "CH"   , "DH"   , "BH"   , 
	"R8B" , "R9B" , "R10B" , "R11B" , "R12B" , "R13B" , "R14B" , "R15B" ,
	// REGS8_REX_BASE : 64
	"SPL" , "BPL" , "SIL"  , "DIL"  ,
	// SREGS_BASE : 68
	"ES"  , "CS"  , "SS"   , "DS"   , "FS"   , "GS"   ,
	// 64-bit PC Register
	"RIP" ,
	// FPUREGS_BASE : 75
	"ST0" , "ST1" , "ST2"  , "ST3"  , "ST4"  , "ST5"  , "ST6"  , "ST7"  ,
	// MMXREGS_BASE : 83
	"MM0",  "MM1",  "MM2" ,  "MM3" ,  "MM4" ,  "MM5" ,  "MM6" ,  "MM7" ,
	// SSEREGS_BASE : 91
	"XMM0", "XMM1", "XMM2" , "XMM3" , "XMM4" , "XMM5" , "XMM6" , "XMM7" , 
	"XMM8", "XMM9", "XMM10", "XMM11", "XMM12", "XMM13", "XMM14", "XMM15",
	// AVXREGS_BASE : 107
	"YMM0", "YMM1", "YMM2" , "YMM3" , "YMM4" , "YMM5" , "YMM6" , "YMM7" , 
	"YMM8", "YMM9", "YMM10", "YMM11", "YMM12", "YMM13", "YMM14", "YMM15",
	// CREGS_BASE : 123
	"CR0" , "CR1" , "CR2"  , "CR3"  , "CR4"  , "CR5"  , "CR6"  , "CR7"  , 
	"CR8" , "CR9" , "CR10" , "CR11" , "CR12" , "CR13" , "CR14" , "CR15" ,
	// DREGS_BASE : 139
	"DR0" , "DR1" , "DR2"  , "DR3"  , "DR4"  , "DR5"  , "DR6"  , "DR7"  ,
	"DR8" , "DR9" , "DR10" , "DR11" , "DR12" , "DR13" , "DR14" , "DR15" 
};

PCSTR PEDisAsm::G_MNEMONICS[] =
{
	"UNKNOWN",			"ADD",				"PUSH",				"POP",
	"OR",				"ADC",				"SBB",				"AND",
	"DAA",				"SUB",				"DAS",				"XOR",
	"AAA",				"CMP",				"AAS",				"INC",
	"DEC",				"PUSHA",			"POPA",				"BOUND",
	"ARPL",				"IMUL",				"INS",				"OUTS",
	"JO",				"JNO",				"JB",				"JNB",
	"JZ",				"JNZ",				"JBE",				"JNBE",
	"JS",				"JNS",				"JP",				"JNP",
	"JL",				"JNL",				"JLE",				"JNLE",
	"TEST",				"XCHG",				"MOV",				"LEA",
	"CBW",				"CWDE",				"CDQE",				"CWD",
	"CDQ",				"CQO",				"CALL FAR",			"PUSHF",
	"POPF",				"SAHF",				"LAHF",				"MOVS",
	"CMPS",				"STOS",				"LODS",				"SCAS",
	"RET",				"LES",				"LDS",				"ENTER",
	"LEAVE",			"RETF",				"INT 3",			"INT",
	"INTO",				"IRET",				"AAM",				"AAD",
	"SALC",				"XLAT",				"LOOPNZ",			"LOOPZ",
	"LOOP",				"JCXZ",				"JECXZ",			"JRCXZ",
	"IN",				"OUT",				"CALL",				"JMP",
	"JMP FAR",			"INT1",				"HLT",				"CMC",
	"CLC",				"STC",				"CLI",				"STI",
	"CLD",				"STD",				"LAR",				"LSL",
	"SYSCALL",			"CLTS",				"SYSRET",			"INVD",
	"WBINVD",			"UD2",				"FEMMS",			"NOP",
	"WRMSR",			"RDTSC",			"RDMSR",			"RDPMC",
	"SYSENTER",			"SYSEXIT",			"GETSEC",			"CMOVO",
	"CMOVNO",			"CMOVB",			"CMOVAE",			"CMOVZ",
	"CMOVNZ",			"CMOVBE",			"CMOVA",			"CMOVS",
	"CMOVNS",			"CMOVP",			"CMOVNP",			"CMOVL",
	"CMOVGE",			"CMOVLE",			"CMOVG",			"SETO",
	"SETNO",			"SETB",				"SETAE",			"SETZ",
	"SETNZ",			"SETBE",			"SETA",				"SETS",
	"SETNS",			"SETP",				"SETNP",			"SETL",
	"SETGE",			"SETLE",			"SETG",				"CPUID",
	"BT",				"SHLD",				"RSM",				"BTS",
	"SHRD",				"CMPXCHG",			"LSS",				"BTR",
	"LFS",				"LGS",				"MOVZX",			"BTC",
	"MOVSX",			"XADD",				"MOVNTI",			"BSWAP",
	"ROL",				"ROR",				"RCL",				"RCR",
	"SHL",				"SHR",				"SAL",				"SAR",
	"FADD",				"FMUL",				"FCOM",				"FCOMP",
	"FSUB",				"FSUBR",			"FDIV",				"FDIVR",
	"FLD",				"FST",				"FSTP",				"FLDENV",
	"FLDCW",			"FXCH",				"FNOP",				"FCHS",
	"FABS",				"FTST",				"FXAM",				"FLD1",
	"FLDL2T",			"FLDL2E",			"FLDPI",			"FLDLG2",
	"FLDLN2",			"FLDZ",				"F2XM1",			"FYL2X",
	"FPTAN",			"FPATAN",			"FXTRACT",			"FPREM1",
	"FDECSTP",			"FINCSTP",			"FPREM",			"FYL2XP1",
	"FSQRT",			"FSINCOS",			"FRNDINT",			"FSCALE",
	"FSIN",				"FCOS",				"FIADD",			"FIMUL",
	"FICOM",			"FICOMP",			"FISUB",			"FISUBR",
	"FIDIV",			"FIDIVR",			"FCMOVB",			"FCMOVE",
	"FCMOVBE",			"FCMOVU",			"FUCOMPP",			"FILD",
	"FISTTP",			"FIST",				"FISTP",			"FCMOVNB",
	"FCMOVNE",			"FCMOVNBE",			"FCMOVNU",			"FENI",
	"FEDISI",			"FSETPM",			"FUCOMI",			"FCOMI",
	"FRSTOR",			"FFREE",			"FUCOM",			"FUCOMP",
	"FADDP",			"FMULP",			"FCOMPP",			"FSUBRP",
	"FSUBP",			"FDIVRP",			"FDIVP",			"FBLD",
	"FBSTP",			"FUCOMIP",			"FCOMIP",			"NOT",
	"NEG",				"MUL",				"DIV",				"IDIV",
	"SLDT",				"STR",				"LLDT",				"LTR",
	"VERR",				"VERW",				"SGDT",				"SIDT",
	"LGDT",				"LIDT",				"SMSW",				"LMSW",
	"INVLPG",			"VMCALL",			"VMLAUNCH",			"VMRESUME",
	"VMXOFF",			"MONITOR",			"MWAIT",			"XGETBV",
	"XSETBV",			"VMFUNC",			"VMRUN",			"VMMCALL",
	"VMLOAD",			"VMSAVE",			"STGI",				"CLGI",
	"SKINIT",			"INVLPGA",			"SWAPGS",			"RDTSCP",
	"PREFETCH",			"PREFETCHW",		"PI2FW",			"PI2FD",
	"PF2IW",			"PF2ID",			"PFNACC",			"PFPNACC",
	"PFCMPGE",			"PFMIN",			"PFRCP",			"PFRSQRT",
	"PFSUB",			"PFADD",			"PFCMPGT",			"PFMAX",
	"PFRCPIT1",			"PFRSQIT1",			"PFSUBR",			"PFACC",
	"PFCMPEQ",			"PFMUL",			"PFRCPIT2",			"PMULHRW",
	"PSWAPD",			"PAVGUSB",			"MOVUPS",			"MOVUPD",
	"MOVSS",			"MOVSD",			"VMOVUPS",			"VMOVUPD",
	"VMOVSS",			"VMOVSD",			"MOVHLPS",			"MOVLPS",
	"MOVLPD",			"MOVSLDUP",			"MOVDDUP",			"VMOVHLPS",
	"VMOVLPS",			"VMOVLPD",			"VMOVSLDUP",		"VMOVDDUP",
	"UNPCKLPS",			"UNPCKLPD",			"VUNPCKLPS",		"VUNPCKLPD",
	"UNPCKHPS",			"UNPCKHPD",			"VUNPCKHPS",		"VUNPCKHPD",
	"MOVLHPS",			"MOVHPS",			"MOVHPD",			"MOVSHDUP",
	"VMOVLHPS",			"VMOVHPS",			"VMOVHPD",			"VMOVSHDUP",
	"PREFETCHNTA",		"PREFETCHT0",		"PREFETCHT1",		"PREFETCHT2",
	"MOVAPS",			"MOVAPD",			"VMOVAPS",			"VMOVAPD",
	"CVTPI2PS",			"CVTPI2PD",			"CVTSI2SS",			"CVTSI2SD",
	"VCVTSI2SS",		"VCVTSI2SD",		"MOVNTPS",			"MOVNTPD",
	"MOVNTSS",			"MOVNTSD",			"VMOVNTPS",			"VMOVNTPD",
	"CVTTPS2PI",		"CVTTPD2PI",		"CVTTSS2SI",		"CVTTSD2SI",
	"VCVTTSS2SI",		"VCVTTSD2SI",		"CVTPS2PI",			"CVTPD2PI",
	"CVTSS2SI",			"CVTSD2SI",			"VCVTSS2SI",		"VCVTSD2SI",
	"UCOMISS",			"UCOMISD",			"VUCOMISS",			"VUCOMISD",
	"COMISS",			"COMISD",			"VCOMISS",			"VCOMISD",
	"MOVMSKPS",			"MOVMSKPD",			"VMOVMSKPS",		"VMOVMSKPD",
	"SQRTPS",			"SQRTPD",			"SQRTSS",			"SQRTSD",
	"VSQRTPS",			"VSQRTPD",			"VSQRTSS",			"VSQRTSD",
	"RSQRTPS",			"RSQRTSS",			"VRSQRTPS",			"VRSQRTSS",
	"RCPPS",			"RCPSS",			"VRCPPS",			"VRCPSS",
	"ANDPS",			"ANDPD",			"VANDPS",			"VANDPD",
	"ANDNPS",			"ANDNPD",			"VANDNPS",			"VANDNPD",
	"ORPS",				"ORPD",				"VORPS",			"VORPD",
	"XORPS",			"XORPD",			"VXORPS",			"VXORPD",
	"ADDPS",			"ADDPD",			"ADDSS",			"ADDSD",
	"VADDPS",			"VADDPD",			"VADDSS",			"VADDSD",
	"MULPS",			"MULPD",			"MULSS",			"MULSD",
	"VMULPS",			"VMULPD",			"VMULSS",			"VMULSD",
	"CVTPS2PD",			"CVTPD2PS",			"CVTSS2SD",			"CVTSD2SS",
	"VCVTPS2PD",		"VCVTPD2PS",		"VCVTSS2SD",		"VCVTSD2SS",
	"CVTDQ2PS",			"CVTPS2DQ",			"CVTTPS2DQ",		"VCVTDQ2PS",
	"VCVTPS2DQ",		"VCVTTPS2DQ",		"SUBPS",			"SUBPD",
	"SUBSS",			"SUBSD",			"VSUBPS",			"VSUBPD",
	"VSUBSS",			"VSUBSD",			"MINPS",			"MINPD",
	"MINSS",			"MINSD",			"VMINPS",			"VMINPD",
	"VMINSS",			"VMINSD",			"DIVPS",			"DIVPD",
	"DIVSS",			"DIVSD",			"VDIVPS",			"VDIVPD",
	"VDIVSS",			"VDIVSD",			"MAXPS",			"MAXPD",
	"MAXSS",			"MAXSD",			"VMAXPS",			"VMAXPD",
	"VMAXSS",			"VMAXSD",			"PUNPCKLBW",		"VPUNPCKLBW",
	"PUNPCKLWD",		"VPUNPCKLWD",		"PUNPCKLDQ",		"VPUNPCKLDQ",
	"PACKSSWB",			"VPACKSSWB",		"PCMPGTB",			"VPCMPGTB",
	"PCMPGTW",			"VPCMPGTW",			"PCMPGTD",			"VPCMPGTD",
	"PACKUSWB",			"VPACKUSWB",		"PUNPCKHBW",		"VPUNPCKHBW",
	"PUNPCKHWD",		"VPUNPCKHWD",		"PUNPCKHDQ",		"VPUNPCKHDQ",
	"PACKSSDW",			"VPACKSSDW",		"PUNPCKLQDQ",		"VPUNPCKLQDQ",
	"PUNPCKHQDQ",		"VPUNPCKHQDQ",		"MOVD",				"MOVQ",
	"VMOVD",			"VMOVQ",			"MOVDQA",			"MOVDQU",
	"VMOVDQA",			"VMOVDQU",			"PSHUFW",			"PSHUFD",
	"PSHUFHW",			"PSHUFLW",			"VPSHUFD",			"VPSHUFHW",
	"VPSHUFLW",			"PCMPEQB",			"VPCMPEQB",			"PCMPEQW",
	"VPCMPEQW",			"PCMPEQD",			"VPCMPEQD",			"EMMS",	
	"VZEROUPPER",		"VZEROALL",			"VMREAD",			"EXTRQ",
	"INSERTQ",			"VMWRITE",			"CVTPH2PS",			"CVTPS2PH",
	"HADDPD",			"HADDPS",			"VHADDPD",			"VHADDPS",
	"HSUBPD",			"HSUBPS",			"VHSUBPD",			"VHSUBPS",
	"XSAVE",			"XSAVE64",			"LFENCE",			"XRSTOR",
	"XRSTOR64",			"MFENCE",			"XSAVEOPT",			"XSAVEOPT64",
	"SFENCE",			"CLFLUSH",			"POPCNT",			"BSF",
	"TZCNT",			"BSR",				"LZCNT",			"CMPEQPS",
	"CMPLTPS",			"CMPLEPS",			"CMPUNORDPS",		"CMPNEQPS",
	"CMPNLTPS",			"CMPNLEPS",			"CMPORDPS",			"CMPEQPD",
	"CMPLTPD",			"CMPLEPD",			"CMPUNORDPD",		"CMPNEQPD",
	"CMPNLTPD",			"CMPNLEPD",			"CMPORDPD",			"CMPEQSS",
	"CMPLTSS",			"CMPLESS",			"CMPUNORDSS",		"CMPNEQSS",
	"CMPNLTSS",			"CMPNLESS",			"CMPORDSS",			"CMPEQSD",
	"CMPLTSD",			"CMPLESD",			"CMPUNORDSD",		"CMPNEQSD",
	"CMPNLTSD",			"CMPNLESD",			"CMPORDSD",			"VCMPEQPS",
	"VCMPLTPS",			"VCMPLEPS",			"VCMPUNORDPS",		"VCMPNEQPS",
	"VCMPNLTPS",		"VCMPNLEPS",		"VCMPORDPS",		"VCMPEQ_UQPS",
	"VCMPNGEPS",		"VCMPNGTPS",		"VCMPFALSEPS",		"VCMPNEQ_OQPS",
	"VCMPGEPS",			"VCMPGTPS",			"VCMPTRUEPS",		"VCMPEQ_OSPS",
	"VCMPLT_OQPS",		"VCMPLE_OQPS",		"VCMPUNORD_SPS",	"VCMPNEQ_USPS",
	"VCMPNLT_UQPS",		"VCMPNLE_UQPS",		"VCMPORD_SPS",		"VCMPEQ_USPS",
	"VCMPNGE_UQPS",		"VCMPNGT_UQPS",		"VCMPFALSE_OSPS",	"VCMPNEQ_OSPS",
	"VCMPGE_OQPS",		"VCMPGT_OQPS",		"VCMPTRUE_USPS",	"VCMPEQPD",
	"VCMPLTPD",			"VCMPLEPD",			"VCMPUNORDPD",		"VCMPNEQPD",
	"VCMPNLTPD",		"VCMPNLEPD",		"VCMPORDPD",		"VCMPEQ_UQPD",
	"VCMPNGEPD",		"VCMPNGTPD",		"VCMPFALSEPD",		"VCMPNEQ_OQPD",
	"VCMPGEPD",			"VCMPGTPD",			"VCMPTRUEPD",		"VCMPEQ_OSPD",
	"VCMPLT_OQPD",		"VCMPLE_OQPD",		"VCMPUNORD_SPD",	"VCMPNEQ_USPD",
	"VCMPNLT_UQPD",		"VCMPNLE_UQPD",		"VCMPORD_SPD",		"VCMPEQ_USPD",
	"VCMPNGE_UQPD",		"VCMPNGT_UQPD",		"VCMPFALSE_OSPD",	"VCMPNEQ_OSPD",
	"VCMPGE_OQPD",		"VCMPGT_OQPD",		"VCMPTRUE_USPD",	"VCMPEQSS",	
	"VCMPLTSS",			"VCMPLESS",			"VCMPUNORDSS",		"VCMPNEQSS",
	"VCMPNLTSS",		"VCMPNLESS",		"VCMPORDSS",		"VCMPEQ_UQSS",
	"VCMPNGESS",		"VCMPNGTSS",		"VCMPFALSESS",		"VCMPNEQ_OQSS",
	"VCMPGESS",			"VCMPGTSS",			"VCMPTRUESS",		"VCMPEQ_OSSS",
	"VCMPLT_OQSS",		"VCMPLE_OQSS",		"VCMPUNORD_SSS",	"VCMPNEQ_USSS",
	"VCMPNLT_UQSS",		"VCMPNLE_UQSS",		"VCMPORD_SSS",		"VCMPEQ_USSS",
	"VCMPNGE_UQSS",		"VCMPNGT_UQSS",		"VCMPFALSE_OSSS",	"VCMPNEQ_OSSS",
	"VCMPGE_OQSS",		"VCMPGT_OQSS",		"VCMPTRUE_USSS",	"VCMPEQSD",
	"VCMPLTSD",			"VCMPLESD",			"VCMPUNORDSD",		"VCMPNEQSD",
	"VCMPNLTSD",		"VCMPNLESD",		"VCMPORDSD",		"VCMPEQ_UQSD",
	"VCMPNGESD",		"VCMPNGTSD",		"VCMPFALSESD",		"VCMPNEQ_OQSD",
	"VCMPGESD",			"VCMPGTSD",			"VCMPTRUESD",		"VCMPEQ_OSSD",
	"VCMPLT_OQSD",		"VCMPLE_OQSD",		"VCMPUNORD_SSD",	"VCMPNEQ_USSD",
	"VCMPNLT_UQSD",		"VCMPNLE_UQSD",		"VCMPORD_SSD",		"VCMPEQ_USSD",
	"VCMPNGE_UQSD",		"VCMPNGT_UQSD",		"VCMPFALSE_OSSD",	"VCMPNEQ_OSSD",
	"VCMPGE_OQSD",		"VCMPGT_OQSD",		"VCMPTRUE_USSD",	"PINSRW",
	"VPINSRW",			"PEXTRW",			"VPEXTRW",			"SHUFPS",
	"SHUFPD",			"VSHUFPS",			"VSHUFPD",			"CMPXCHG8B",
	"CMPXCHG16B",		"VMPTRST",			"ADDSUBPD",			"ADDSUBPS",
	"VADDSUBPD",		"VADDSUBPS",		"PSRLW",			"VPSRLW",
	"PSRLD",			"VPSRLD",			"PSRLQ",			"VPSRLQ",
	"PADDQ",			"VPADDQ",			"PMULLW",			"VPMULLW",
	"MOVQ2DQ",			"MOVDQ2Q",			"PMOVMSKB",			"VPMOVMSKB",
	"PSUBUSB",			"VPSUBUSB",			"PSUBUSW",			"VPSUBUSW",
	"PMINUB",			"VPMINUB",			"PAND",				"VPAND",
	"PADDUSB",			"VPADDUSW",			"PADDUSW",			"PMAXUB",
	"VPMAXUB",			"PANDN",			"VPANDN",			"PAVGB",
	"VPAVGB",			"PSRAW",			"VPSRAW",			"PSRAD",
	"VPSRAD",			"PAVGW",			"VPAVGW",			"PMULHUW",
	"VPMULHUW",			"PMULHW",			"VPMULHW",			"CVTTPD2DQ",
	"CVTDQ2PD",			"CVTPD2DQ",			"VCVTTPD2DQ",		"VCVTDQ2PD",
	"VCVTPD2DQ",		"MOVNTQ",			"MOVNTDQ",			"VMOVNTDQ",
	"PSUBSB",			"VPSUBSB",			"PSUBSW",			"VPSUBSW",
	"PMINSW",			"VPMINSW",			"POR",				"VPOR",
	"PADDSB",			"VPADDSB",			"PADDSW",			"VPADDSW",
	"PMAXSW",			"VPMAXSW",			"PXOR",				"VPXOR",
	"LDDQU",			"VLDDQU",			"PSLLW",			"VPSLLW",
	"PSLLD",			"VPSLLD",			"PSLLQ",			"VPSLLQ",
	"PMULUDQ",			"VPMULUDQ",			"PMADDWD",			"VPMADDWD",
	"PSADBW",			"VPSADBW",			"MASKMOVQ",			"MASKMOVDQU",
	"VMASKMOVDQU",		"PSUBB",			"VPSUBB",			"PSUBW",
	"VPSUBW",			"PSUBD",			"VPSUBD",			"PSUBQ",
	"VPSUBQ",			"PADDB",			"VPADDB",			"PADDW",
	"VPADDW",			"PADDD",			"VPADDD",			"FNSTENV",
	"FSTENV",			"FNSTCW",			"FSTCW",			"FNCLEX",
	"FCLEX",			"FNINIT",			"FINIT",			"FNSAVE",
	"FSAVE",			"FNSTSW",			"FSTSW",			"PSHUFB",
	"VPSHUFB",			"PHADDW",			"VPHADDW",			"PHADDD",
	"VPHADDD",			"PHADDSW",			"VPHADDSW",			"PMADDUBSW",
	"VPMADDUBSW",		"PHSUBW",			"VPHSUBW",			"PHSUBD",
	"VPHSUBD",			"PHSUBSW",			"VPHSUBSW",			"PSIGNB",
	"VPSIGNB",			"PSIGNW",			"VPSIGNW",			"PSIGND",
	"VPSIGND",			"PMULHRSW",			"VPMULHRSW",		"VPERMILPS",
	"VPERMILPD",		"VTESTPS",			"VTESTPD",			"PBLENDVB",
	"BLENDVPS",			"BLENDVPD",			"PTEST",			"VPTEST",
	"VBROADCASTSS",		"VBROADCASTSD",		"VBROADCASTF128",	"PABSB",
	"VPABSB",			"PABSW",			"VPABSW",			"PABSD",
	"VPABSD",			"PMOVSXBW",			"VPMOVSXBW",		"PMOVSXBD",
	"VPMOVSXBD",		"PMOVSXBQ",			"VPMOVSXBQ",		"PMOVSXWD",
	"VPMOVSXWD",		"PMOVSXWQ",			"VPMOVSXWQ",		"PMOVSXDQ",
	"VPMOVSXDQ",		"PMULDQ",			"VPMULDQ",			"PCMPEQQ",
	"VPCMPEQQ",			"MOVNTDQA",			"VMOVNTDQA",		"PACKUSDW",
	"VPACKUSDW",		"VMASKMOVPS",		"VMASKMOVPD",		"PMOVZXBW",
	"VPMOVZXBW",		"PMOVZXBD",			"VPMOVZXBD",		"PMOVZXBQ",
	"VPMOVZXBQ",		"PMOVZXWD",			"VPMOVZXWD",		"PMOVZXWQ",
	"VPMOVZXWQ",		"PMOVZXDQ",			"VPMOVZXDQ",		"PCMPGTQ",
	"VPCMPGTQ",			"PMINSB",			"VPMINSB",			"PMINSD",
	"VPMINSD",			"PMINUW",			"VPMINUW",			"PMINUD",
	"VPMINUD",			"PMAXSB",			"VPMAXSB",			"PMAXSD",
	"VPMAXSD",			"PMAXUW",			"VPMAXUW",			"PMAXUD",
	"VPMAXUD",			"PMULLD",			"VPMULLD",			"PHMINPOSUW",
	"VPHMINPOSUW",		"INVEPT",			"INVVPID",			"INVPCID",
	"VFMADDSUB132PS",	"VFMADDSUB132PD",	"VFMSUBADD132PS",	"VFMSUBADD132PD",
	"VFMADD132PS",		"VFMADD132PD",		"VFMADD132SS",		"VFMADD132SD",
	"VFMSUB132PS",		"VFMSUB132PD",		"VFMSUB132SS",		"VFMSUB132SD",
	"VFNMADD132PS",		"VFNMADD132PD",		"VFNMADD132SS",		"VFNMADD132SD",
	"VFNMSUB132PS",		"VFNMSUB132PD",		"VFNMSUB132SS",		"VFNMSUB132SD",
	"VFMADDSUB213PS",	"VFMADDSUB213PD",	"VFMSUBADD213PS",	"VFMSUBADD213PD",
	"VFMADD213PS",		"VFMADD213PD",		"VFMADD213SS",		"VFMADD213SD",
	"VFMSUB213PS",		"VFMSUB213PD",		"VFMSUB213SS",		"VFMSUB213SD",
	"VFNMADD213PS",		"VFNMADD213PD",		"VFNMADD213SS",		"VFNMADD213SD",
	"VFNMSUB213PS",		"VFNMSUB213PD",		"VFNMSUB213SS",		"VFNMSUB213SD",
	"VFMADDSUB231PS",	"VFMADDSUB231PD",	"VFMSUBADD231PS",	"VFMSUBADD231PD",
	"VFMADD231PS",		"VFMADD231PD",		"VFMADD231SS",		"VFMADD231SD",
	"VFMSUB231PS",		"VFMSUB231PD",		"VFMSUB231SS",		"VFMSUB231SD",
	"VFNMADD231PS",		"VFNMADD231PD",		"VFNMADD231SS",		"VFNMADD231SD",
	"VFNMSUB231PS",		"VFNMSUB231PD",		"VFNMSUB231SS",		"VFNMSUB231SD",
	"AESIMC",			"VAESIMC",			"AESENC",			"VAESENC",
	"AESENCLAST",		"VAESENCLAST",		"AESDEC",			"VAESDEC",
	"AESDECLAST",		"VAESDECLAST",		"MOVBE",			"CRC32",
	"VPERM2F128",		"ROUNDPS",			"VROUNDPS",			"ROUNDPD",
	"VROUNDPD",			"ROUNDSS",			"VROUNDSS",			"ROUNDSD",
	"VROUNDSD",			"BLENDPS",			"VBLENDPS",			"BLENDPD",
	"VBLENDPD",			"PBLENDW",			"VPBLENDW",			"PALIGNR",
	"VPALIGNR",			"PEXTRB",			"VPEXTRB",			"PEXTRD",
	"PEXTRQ",			"VPEXTRD",			"VPEXTRQ",			"EXTRACTPS",
	"VEXTRACTPS",		"VINSERTF128",		"VEXTRACTF128",		"PINSRB",
	"VPINSRB",			"INSERTPS",			"VINSERTPS",		"PINSRD",
	"PINSRQ",			"VPINSRD",			"VPINSRQ",			"DPPS",
	"VDPPS",			"DPPD",				"VDPPD",			"MPSADBW",
	"VMPSADBW",			"PCLMULQDQ",		"VPCLMULQDQ",		"VBLENDVPS",
	"VBLENDVPD",		"VPBLENDVB",		"PCMPESTRM",		"VPCMPESTRM",
	"PCMPESTRI",		"VPCMPESTRI",		"PCMPISTRM",		"VPCMPISTRM",
	"PCMPISTRI",		"VPCMPISTRI",		"AESKEYGENASSIST",	"VAESKEYGENASSIST",
	"PSRLDQ",			"VPSRLDQ",			"PSLLDQ",			"VPSLLDQ",
	"FXSAVE",			"FXSAVE64",			"RDFSBASE",			"FXRSTOR",
	"FXRSTOR64",		"RDGSBASE",			"LDMXCSR",			"WRFSBASE",
	"VLDMXCSR",			"STMXCSR",			"WRGSBASE",			"VSTMXCSR",
	"VMPTRLD",			"VMCLEAR",			"VMXON",			"MOVSXD",
	"PAUSE",			"WAIT",				"RDRAND",			"RDSEED",
	//-------------------------------------------------------------------------------//
	"VPERMQ",			"VPERMPD",			"VPBLENDD",			"VPERM2I128",
	"VCVTPS2PH",		"VCVTPH2PS",		"VINSERTI128",		"VEXTRACTI128",
	"RORX",				"VPERMPS",			"VPERMD",			"VPSRLVD",
	"VPSRAVD",			"VPSLLVD",			"VPSLLVQ",			"VPBROADCASTD",
	"VPBROADCASTQ",		"VPSRLVQ",			"VBROADCASTI128",	"VPBROADCASTB",
	"VPBROADCASTW",		"VPGATHERDD",		"VPGATHERDQ",		"VPGATHERQD",
	"VPGATHERQQ",		"VGATHERDPS",		"VGATHERDPD",		"VGATHERQPS",
	"VGATHERQPD",		"VPMASKMOVD",		"VPMASKMOVQ",		"ANDN",
	"BZHI",				"BEXTR",			"PEXT",				"PDEP",
	"ADCX",				"ADOX",				"MULX",				"SHLX",
	"SARX",				"SHRX",				"BLSR",				"BLSMSK",
	"BLSI"
};


OP_INFO PEDisAsm::OP_INFOSPS[5] = 
{
	{ 0x08, OI_NOP, OT_NONE, OT_NONE, 0x00000000 },
	{ 0x08, OI_PAUSE, OT_NONE, OT_NONE, 0x00000001 },
	{ 0x08, OI_MOVSXD, OT_REG32_64, OT_RM32, 0x84002800 },
};

OP_INFO PEDisAsm::OP_INFOS[OP_INFO_SIZE] =
{ //1251
	/*0x0000*/{ 0x08, OI_ADD, OT_RM8, OT_REG8, 0x80002001 },
	/*0x0001*/{ 0x08, OI_ADD, OT_RM_ALL, OT_REG_ALL, 0x80002001 },
	/*0x0002*/{ 0x08, OI_ADD, OT_REG8, OT_RM8, 0x80002001 },
	/*0x0003*/{ 0x08, OI_ADD, OT_REG_ALL, OT_RM_ALL, 0x80002001 },
	/*0x0004*/{ 0x08, OI_ADD, OT_ACC8, OT_IMM8, 0x80000000 },
	/*0x0005*/{ 0x08, OI_ADD, OT_ACC_ALL, OT_IMM_ALL, 0x80000000 },
	/*0x0006*/{ 0x08, OI_PUSH, OT_SEG_PF, OT_NONE, 0x02000040 },
	/*0x0007*/{ 0x08, OI_POP, OT_SEG_PF, OT_NONE, 0x82000040 },
	/*0x0008*/{ 0x08, OI_OR, OT_RM8, OT_REG8, 0x80002001 },
	/*0x0009*/{ 0x08, OI_OR, OT_RM_ALL, OT_REG_ALL, 0x80002001 },
	/*0x000A*/{ 0x08, OI_OR, OT_REG8, OT_RM8, 0x80002001 },
	/*0x000B*/{ 0x08, OI_OR, OT_REG_ALL, OT_RM_ALL, 0x80002001 },
	/*0x000C*/{ 0x08, OI_OR, OT_ACC8, OT_IMM8, 0x80000000 },
	/*0x000D*/{ 0x08, OI_OR, OT_ACC_ALL, OT_IMM_ALL, 0x80000000 },
	/*0x000E*/{ 0x08, OI_PUSH, OT_SEG_PF, OT_NONE, 0x02000008 },
	/*0x000F*/{ 0x08, OI_ADC, OT_RM8, OT_REG8, 0x80002001 },
	/*0x0010*/{ 0x08, OI_ADC, OT_RM_ALL, OT_REG_ALL, 0x80002001 },
	/*0x0011*/{ 0x08, OI_ADC, OT_REG8, OT_RM8, 0x80002001 },
	/*0x0012*/{ 0x08, OI_ADC, OT_REG_ALL, OT_RM_ALL, 0x80002001 },
	/*0x0013*/{ 0x08, OI_ADC, OT_ACC8, OT_IMM8, 0x80000000 },
	/*0x0014*/{ 0x08, OI_ADC, OT_ACC_ALL, OT_IMM_ALL, 0x80000000 },
	/*0x0015*/{ 0x08, OI_PUSH, OT_SEG_PF, OT_NONE, 0x02000010 },
	/*0x0016*/{ 0x08, OI_POP, OT_SEG_PF, OT_NONE, 0x82000010 },
	/*0x0017*/{ 0x08, OI_SBB, OT_RM8, OT_REG8, 0x80002001 },
	/*0x0018*/{ 0x08, OI_SBB, OT_RM_ALL, OT_REG_ALL, 0x80002001 },
	/*0x0019*/{ 0x08, OI_SBB, OT_REG8, OT_RM8, 0x80002001 },
	/*0x001A*/{ 0x08, OI_SBB, OT_REG_ALL, OT_RM_ALL, 0x80002001 },
	/*0x001B*/{ 0x08, OI_SBB, OT_ACC8, OT_IMM8, 0x80000000 },
	/*0x001C*/{ 0x08, OI_SBB, OT_ACC_ALL, OT_IMM_ALL, 0x80000000 },
	/*0x001D*/{ 0x08, OI_PUSH, OT_SEG_PF, OT_NONE, 0x02000020 },
	/*0x001E*/{ 0x08, OI_POP, OT_SEG_PF, OT_NONE, 0x82000020 },
	/*0x001F*/{ 0x08, OI_AND, OT_RM8, OT_REG8, 0x80002001 },
	/*0x0020*/{ 0x08, OI_AND, OT_RM_ALL, OT_REG_ALL, 0x80002001 },
	/*0x0021*/{ 0x08, OI_AND, OT_REG8, OT_RM8, 0x80002001 },
	/*0x0022*/{ 0x08, OI_AND, OT_REG_ALL, OT_RM_ALL, 0x80002001 },
	/*0x0023*/{ 0x08, OI_AND, OT_ACC8, OT_IMM8, 0x80000000 },
	/*0x0024*/{ 0x08, OI_AND, OT_ACC_ALL, OT_IMM_ALL, 0x80000000 },
	/*0x0025*/{ 0x08, OI_DAA, OT_NONE, OT_NONE, 0x02000000 },
	/*0x0026*/{ 0x08, OI_SUB, OT_RM8, OT_REG8, 0x80002001 },
	/*0x0027*/{ 0x08, OI_SUB, OT_RM_ALL, OT_REG_ALL, 0x80002001 },
	/*0x0028*/{ 0x08, OI_SUB, OT_REG8, OT_RM8, 0x80002001 },
	/*0x0029*/{ 0x08, OI_SUB, OT_REG_ALL, OT_RM_ALL, 0x80002001 },
	/*0x002A*/{ 0x08, OI_SUB, OT_ACC8, OT_IMM8, 0x80000000 },
	/*0x002B*/{ 0x08, OI_SUB, OT_ACC_ALL, OT_IMM_ALL, 0x80000000 },
	/*0x002C*/{ 0x08, OI_DAS, OT_NONE, OT_NONE, 0x02000000 },
	/*0x002D*/{ 0x08, OI_XOR, OT_RM8, OT_REG8, 0x80002001 },
	/*0x002E*/{ 0x08, OI_XOR, OT_RM_ALL, OT_REG_ALL, 0x80002001 },
	/*0x002F*/{ 0x08, OI_XOR, OT_REG8, OT_RM8, 0x80002001 },
	/*0x0030*/{ 0x08, OI_XOR, OT_REG_ALL, OT_RM_ALL, 0x80002001 },
	/*0x0031*/{ 0x08, OI_XOR, OT_ACC8, OT_IMM8, 0x80000000 },
	/*0x0032*/{ 0x08, OI_XOR, OT_ACC_ALL, OT_IMM_ALL, 0x80000000 },
	/*0x0033*/{ 0x08, OI_AAA, OT_NONE, OT_NONE, 0x02000000 },
	/*0x0034*/{ 0x08, OI_CMP, OT_RM8, OT_REG8, 0x00002000 },
	/*0x0035*/{ 0x08, OI_CMP, OT_RM_ALL, OT_REG_ALL, 0x00002000 },
	/*0x0036*/{ 0x08, OI_CMP, OT_REG8, OT_RM8, 0x00002000 },
	/*0x0037*/{ 0x08, OI_CMP, OT_REG_ALL, OT_RM_ALL, 0x00002000 },
	/*0x0038*/{ 0x08, OI_CMP, OT_ACC8, OT_IMM8, 0x00000000 },
	/*0x0039*/{ 0x08, OI_CMP, OT_ACC_ALL, OT_IMM_ALL, 0x00000000 },
	/*0x003A*/{ 0x08, OI_AAS, OT_NONE, OT_NONE, 0x02000000 },
	/*0x003B*/{ 0x08, OI_INC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x003C*/{ 0x08, OI_INC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x003D*/{ 0x08, OI_INC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x003E*/{ 0x08, OI_INC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x003F*/{ 0x08, OI_INC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x0040*/{ 0x08, OI_INC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x0041*/{ 0x08, OI_INC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x0042*/{ 0x08, OI_INC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x0043*/{ 0x08, OI_DEC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x0044*/{ 0x08, OI_DEC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x0045*/{ 0x08, OI_DEC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x0046*/{ 0x08, OI_DEC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x0047*/{ 0x08, OI_DEC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x0048*/{ 0x08, OI_DEC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x0049*/{ 0x08, OI_DEC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x004A*/{ 0x08, OI_DEC, OT_OP_R_ALL, OT_NONE, 0x82000000 },
	/*0x004B*/{ 0x08, OI_PUSH, OT_OP_R_ALL, OT_NONE, 0x04000000 },
	/*0x004C*/{ 0x08, OI_PUSH, OT_OP_R_ALL, OT_NONE, 0x04000000 },
	/*0x004D*/{ 0x08, OI_PUSH, OT_OP_R_ALL, OT_NONE, 0x04000000 },
	/*0x004E*/{ 0x08, OI_PUSH, OT_OP_R_ALL, OT_NONE, 0x04000000 },
	/*0x004F*/{ 0x08, OI_PUSH, OT_OP_R_ALL, OT_NONE, 0x04000000 },
	/*0x0050*/{ 0x08, OI_PUSH, OT_OP_R_ALL, OT_NONE, 0x04000000 },
	/*0x0051*/{ 0x08, OI_PUSH, OT_OP_R_ALL, OT_NONE, 0x04000000 },
	/*0x0052*/{ 0x08, OI_PUSH, OT_OP_R_ALL, OT_NONE, 0x04000000 },
	/*0x0053*/{ 0x08, OI_POP, OT_OP_R_ALL, OT_NONE, 0x84000000 },
	/*0x0054*/{ 0x08, OI_POP, OT_OP_R_ALL, OT_NONE, 0x84000000 },
	/*0x0055*/{ 0x08, OI_POP, OT_OP_R_ALL, OT_NONE, 0x84000000 },
	/*0x0056*/{ 0x08, OI_POP, OT_OP_R_ALL, OT_NONE, 0x84000000 },
	/*0x0057*/{ 0x08, OI_POP, OT_OP_R_ALL, OT_NONE, 0x84000000 },
	/*0x0058*/{ 0x08, OI_POP, OT_OP_R_ALL, OT_NONE, 0x84000000 },
	/*0x0059*/{ 0x08, OI_POP, OT_OP_R_ALL, OT_NONE, 0x84000000 },
	/*0x005A*/{ 0x08, OI_POP, OT_OP_R_ALL, OT_NONE, 0x84000000 },
	/*0x005B*/{ 0x08, OI_PUSHA, OT_NONE, OT_NONE, 0x02020000 },
	/*0x005C*/{ 0x08, OI_POPA, OT_NONE, OT_NONE, 0x02020000 },
	/*0x005D*/{ 0x08, OI_BOUND, OT_REG_ALL, OT_M1632_A_1632, 0x02002000 },
	/*0x005E*/{ 0x08, OI_ARPL, OT_RM16, OT_REG16, 0x00002000 },
	/*0x005F*/{ 0x08, OI_PUSH, OT_IMM_ALL, OT_NONE, 0x04000000 },
	/*0x0060*/{ 0x08, OI_PUSH, OT_SE_IMM8, OT_NONE, 0x04000200 },
	/*0x0061*/{ 0x08, OI_INS, OT_REG_EDI, OT_REG_DX, 0x00000006 },
	/*0x0062*/{ 0x08, OI_INS, OT_REG_EDI, OT_REG_DX, 0x00008006 },
	/*0x0063*/{ 0x08, OI_OUTS, OT_REG_DX, OT_REG_ESI, 0x00000006 },
	/*0x0064*/{ 0x08, OI_OUTS, OT_REG_DX, OT_REG_ESI, 0x00008006 },
	/*0x0065*/{ 0x0D, OI_JO, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x0066*/{ 0x0D, OI_JNO, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x0067*/{ 0x0D, OI_JB, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x0068*/{ 0x0D, OI_JNB, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x0069*/{ 0x0D, OI_JZ, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x006A*/{ 0x0D, OI_JNZ, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x006B*/{ 0x0D, OI_JBE, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x006C*/{ 0x0D, OI_JNBE, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x006D*/{ 0x0D, OI_JS, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x006E*/{ 0x0D, OI_JNS, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x006F*/{ 0x0D, OI_JP, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x0070*/{ 0x0D, OI_JNP, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x0071*/{ 0x0D, OI_JL, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x0072*/{ 0x0D, OI_JNL, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x0073*/{ 0x0D, OI_JLE, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x0074*/{ 0x0D, OI_JNLE, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x0075*/{ 0x08, OI_TEST, OT_RM8, OT_REG8, 0x00002000 },
	/*0x0076*/{ 0x08, OI_TEST, OT_RM_ALL, OT_REG_ALL, 0x00002000 },
	/*0x0077*/{ 0x08, OI_XCHG, OT_RM8, OT_REG8, 0x80002001 },
	/*0x0078*/{ 0x08, OI_XCHG, OT_RM_ALL, OT_REG_ALL, 0x80002001 },
	/*0x0079*/{ 0x08, OI_MOV, OT_RM8, OT_REG8, 0x80002000 },
	/*0x007A*/{ 0x08, OI_MOV, OT_RM_ALL, OT_REG_ALL, 0x80002000 },
	/*0x007B*/{ 0x08, OI_MOV, OT_REG8, OT_RM8, 0x80002000 },
	/*0x007C*/{ 0x08, OI_MOV, OT_REG_ALL, OT_RM_ALL, 0x80002000 },
	/*0x007D*/{ 0x08, OI_MOV, OT_R_ALL_M16, OT_REG_SEG, 0x80002000 },
	/*0x007E*/{ 0x08, OI_LEA, OT_REG_ALL, OT_MEM, 0x80002000 },
	/*0x007F*/{ 0x08, OI_MOV, OT_REG_SEG, OT_R_ALL_M16, 0x80002000 },
	/*0x0080*/{ 0x08, OI_XCHG, OT_OP_R_ALL, OT_ACC_ALL, 0x80000000 },
	/*0x0081*/{ 0x08, OI_XCHG, OT_OP_R_ALL, OT_ACC_ALL, 0x80000000 },
	/*0x0082*/{ 0x08, OI_XCHG, OT_OP_R_ALL, OT_ACC_ALL, 0x80000000 },
	/*0x0083*/{ 0x08, OI_XCHG, OT_OP_R_ALL, OT_ACC_ALL, 0x80000000 },
	/*0x0084*/{ 0x08, OI_XCHG, OT_OP_R_ALL, OT_ACC_ALL, 0x80000000 },
	/*0x0085*/{ 0x08, OI_XCHG, OT_OP_R_ALL, OT_ACC_ALL, 0x80000000 },
	/*0x0086*/{ 0x08, OI_XCHG, OT_OP_R_ALL, OT_ACC_ALL, 0x80000000 },
	/*0x0087*/{ 0x08, OI_XCHG, OT_OP_R_ALL, OT_ACC_ALL, 0x80000000 },
	/*0x0088*/{ 0x09, OI_CALL_FAR, OT_PTR16_ALL, OT_NONE, 0x02000000 },
	/*0x0089*/{ 0x08, OI_PUSHF, OT_NONE, OT_NONE, 0x04020000 },
	/*0x008A*/{ 0x08, OI_POPF, OT_NONE, OT_NONE, 0x04020000 },
	/*0x008B*/{ 0x08, OI_SAHF, OT_NONE, OT_NONE, 0x00000000 },
	/*0x008C*/{ 0x08, OI_LAHF, OT_NONE, OT_NONE, 0x00000000 },
	/*0x008D*/{ 0x08, OI_MOV, OT_ACC8, OT_MOFFS8, 0x80000000 },
	/*0x008E*/{ 0x08, OI_MOV, OT_ACC_ALL, OT_MOFFS_ALL, 0x80000000 },
	/*0x008F*/{ 0x08, OI_MOV, OT_MOFFS8, OT_ACC8, 0x80000000 },
	/*0x0090*/{ 0x08, OI_MOV, OT_MOFFS_ALL, OT_ACC_ALL, 0x80000000 },
	/*0x0091*/{ 0x08, OI_MOVS, OT_REG_EDI, OT_REG_ESI, 0x80000006 },
	/*0x0092*/{ 0x08, OI_MOVS, OT_REG_EDI, OT_REG_ESI, 0x84008806 },
	/*0x0093*/{ 0x08, OI_CMPS, OT_REG_ESI, OT_REG_EDI, 0x00000006 },
	/*0x0094*/{ 0x08, OI_CMPS, OT_REG_ESI, OT_REG_EDI, 0x04008806 },
	/*0x0095*/{ 0x08, OI_TEST, OT_ACC8, OT_IMM8, 0x00000000 },
	/*0x0096*/{ 0x08, OI_TEST, OT_ACC_ALL, OT_IMM_ALL, 0x00000000 },
	/*0x0097*/{ 0x08, OI_STOS, OT_REG_EDI, OT_ACC8, 0x00000006 },
	/*0x0098*/{ 0x08, OI_STOS, OT_REG_EDI, OT_ACC_ALL, 0x04008806 },
	/*0x0099*/{ 0x08, OI_LODS, OT_ACC8, OT_REG_ESI, 0x00000006 },
	/*0x009A*/{ 0x08, OI_LODS, OT_ACC_ALL, OT_REG_ESI, 0x04008806 },
	/*0x009B*/{ 0x08, OI_SCAS, OT_REG_EDI, OT_ACC8, 0x00000006 },
	/*0x009C*/{ 0x08, OI_SCAS, OT_REG_EDI, OT_ACC_ALL, 0x04008806 },
	/*0x009D*/{ 0x08, OI_MOV, OT_OP_R8, OT_IMM8, 0x80000000 },
	/*0x009E*/{ 0x08, OI_MOV, OT_OP_R8, OT_IMM8, 0x80000000 },
	/*0x009F*/{ 0x08, OI_MOV, OT_OP_R8, OT_IMM8, 0x80000000 },
	/*0x00A0*/{ 0x08, OI_MOV, OT_OP_R8, OT_IMM8, 0x80000000 },
	/*0x00A1*/{ 0x08, OI_MOV, OT_OP_R8, OT_IMM8, 0x80000000 },
	/*0x00A2*/{ 0x08, OI_MOV, OT_OP_R8, OT_IMM8, 0x80000000 },
	/*0x00A3*/{ 0x08, OI_MOV, OT_OP_R8, OT_IMM8, 0x80000000 },
	/*0x00A4*/{ 0x08, OI_MOV, OT_OP_R8, OT_IMM8, 0x80000000 },
	/*0x00A5*/{ 0x08, OI_MOV, OT_OP_R_ALL, OT_IMM_ALL, 0x84000800 },
	/*0x00A6*/{ 0x08, OI_MOV, OT_OP_R_ALL, OT_IMM_ALL, 0x84000800 },
	/*0x00A7*/{ 0x08, OI_MOV, OT_OP_R_ALL, OT_IMM_ALL, 0x84000800 },
	/*0x00A8*/{ 0x08, OI_MOV, OT_OP_R_ALL, OT_IMM_ALL, 0x84000800 },
	/*0x00A9*/{ 0x08, OI_MOV, OT_OP_R_ALL, OT_IMM_ALL, 0x84000800 },
	/*0x00AA*/{ 0x08, OI_MOV, OT_OP_R_ALL, OT_IMM_ALL, 0x84000800 },
	/*0x00AB*/{ 0x08, OI_MOV, OT_OP_R_ALL, OT_IMM_ALL, 0x84000800 },
	/*0x00AC*/{ 0x08, OI_MOV, OT_OP_R_ALL, OT_IMM_ALL, 0x84000800 },
	/*0x00AD*/{ 0x0A, OI_RET, OT_IMM16, OT_NONE, 0x04000000 },
	/*0x00AE*/{ 0x0A, OI_RET, OT_NONE, OT_NONE, 0x04000000 },
	/*0x00AF*/{ 0x08, OI_LES, OT_REG_ALL, OT_MEM16_ALL, 0x02002000 },
	/*0x00B0*/{ 0x08, OI_LDS, OT_REG_ALL, OT_MEM16_ALL, 0x02002000 },
	/*0x00B1*/{ 0x08, OI_ENTER, OT_IMM16, OT_IMM8, 0x04000000 },
	/*0x00B2*/{ 0x08, OI_LEAVE, OT_NONE, OT_NONE, 0x04000000 },
	/*0x00B3*/{ 0x0A, OI_RETF, OT_IMM16, OT_NONE, 0x04020800 },
	/*0x00B4*/{ 0x0A, OI_RETF, OT_NONE, OT_NONE, 0x04020800 },
	/*0x00B5*/{ 0x0E, OI_INT_3, OT_NONE, OT_NONE, 0x00000000 },
	/*0x00B6*/{ 0x0E, OI_INT, OT_IMM8, OT_NONE, 0x00000000 },
	/*0x00B7*/{ 0x0E, OI_INTO, OT_NONE, OT_NONE, 0x02000000 },
	/*0x00B8*/{ 0x0A, OI_IRET, OT_NONE, OT_NONE, 0x04020800 },
	/*0x00B9*/{ 0x08, OI_AAM, OT_IMM8, OT_NONE, 0x02000000 },
	/*0x00BA*/{ 0x08, OI_AAD, OT_IMM8, OT_NONE, 0x02000000 },
	/*0x00BB*/{ 0x08, OI_SALC, OT_NONE, OT_NONE, 0x02000000 },
	/*0x00BC*/{ 0x08, OI_XLAT, OT_REGI_EBXAL, OT_NONE, 0x00000020 },
	/*0x00BD*/{ 0x0D, OI_LOOPNZ, OT_REL_8, OT_NONE, 0x00020400 },
	/*0x00BE*/{ 0x0D, OI_LOOPZ, OT_REL_8, OT_NONE, 0x00020400 },
	/*0x00BF*/{ 0x0D, OI_LOOP, OT_REL_8, OT_NONE, 0x00020400 },
	/*0x00C0*/{ 0x08, OI_IN, OT_ACC8, OT_IMM8, 0x80000000 },
	/*0x00C1*/{ 0x08, OI_IN, OT_ACC_ALL_NO64, OT_IMM8, 0x80000000 },
	/*0x00C2*/{ 0x08, OI_OUT, OT_IMM8, OT_ACC8, 0x00000000 },
	/*0x00C3*/{ 0x08, OI_OUT, OT_IMM8, OT_ACC_ALL_NO64, 0x00000000 },
	/*0x00C4*/{ 0x09, OI_CALL, OT_REL_16_32, OT_NONE, 0x04000000 },
	/*0x00C5*/{ 0x0C, OI_JMP, OT_REL_16_32, OT_NONE, 0x04000000 },
	/*0x00C6*/{ 0x0C, OI_JMP_FAR, OT_PTR16_ALL, OT_NONE, 0x02000000 },
	/*0x00C7*/{ 0x0C, OI_JMP, OT_REL_8, OT_NONE, 0x04000000 },
	/*0x00C8*/{ 0x08, OI_IN, OT_ACC8, OT_REG_DX, 0x80000000 },
	/*0x00C9*/{ 0x08, OI_IN, OT_ACC_ALL_NO64, OT_REG_DX, 0x80000000 },
	/*0x00CA*/{ 0x08, OI_OUT, OT_REG_DX, OT_ACC8, 0x00000000 },
	/*0x00CB*/{ 0x08, OI_OUT, OT_REG_DX, OT_ACC_ALL_NO64, 0x00000000 },
	/*0x00CC*/{ 0x0E, OI_INT1, OT_NONE, OT_NONE, 0x00000000 },
	/*0x00CD*/{ 0x08, OI_HLT, OT_NONE, OT_NONE, 0x00000000 },
	/*0x00CE*/{ 0x08, OI_CMC, OT_NONE, OT_NONE, 0x00000000 },
	/*0x00CF*/{ 0x08, OI_CLC, OT_NONE, OT_NONE, 0x00000000 },
	/*0x00D0*/{ 0x08, OI_STC, OT_NONE, OT_NONE, 0x00000000 },
	/*0x00D1*/{ 0x08, OI_CLI, OT_NONE, OT_NONE, 0x00000000 },
	/*0x00D2*/{ 0x08, OI_STI, OT_NONE, OT_NONE, 0x00000000 },
	/*0x00D3*/{ 0x08, OI_CLD, OT_NONE, OT_NONE, 0x00000000 },
	/*0x00D4*/{ 0x08, OI_STD, OT_NONE, OT_NONE, 0x00000000 },
	/*0x00D5*/{ 0x08, OI_LAR, OT_REG_ALL, OT_RM16, 0x00002000 },
	/*0x00D6*/{ 0x08, OI_LSL, OT_REG_ALL, OT_RM16, 0x00002000 },
	/*0x00D7*/{ 0x1B, OI_SYSCALL, OT_NONE, OT_NONE, 0x00010000 },
	/*0x00D8*/{ 0x08, OI_CLTS, OT_NONE, OT_NONE, 0x00010000 },
	/*0x00D9*/{ 0x1B, OI_SYSRET, OT_NONE, OT_NONE, 0x00010000 },
	/*0x00DA*/{ 0x08, OI_INVD, OT_NONE, OT_NONE, 0x00010000 },
	/*0x00DB*/{ 0x08, OI_WBINVD, OT_NONE, OT_NONE, 0x00010000 },
	/*0x00DC*/{ 0x0E, OI_UD2, OT_NONE, OT_NONE, 0x00010000 },
	/*0x00DD*/{ 0x60, OI_FEMMS, OT_NONE, OT_NONE, 0x00000000 },
	/*0x00DE*/{ 0x08, OI_NOP, OT_RM_ALL, OT_NONE, 0x00002000 },
	/*0x00DF*/{ 0x08, OI_MOV, OT_REG32_64_RM, OT_REG_CR, 0x84012000 },
	/*0x00E0*/{ 0x08, OI_MOV, OT_REG32_64_RM, OT_REG_DR, 0x84012000 },
	/*0x00E1*/{ 0x08, OI_MOV, OT_REG_CR, OT_REG32_64_RM, 0x84012000 },
	/*0x00E2*/{ 0x08, OI_MOV, OT_REG_DR, OT_REG32_64_RM, 0x84012000 },
	/*0x00E3*/{ 0x08, OI_WRMSR, OT_NONE, OT_NONE, 0x00010000 },
	/*0x00E4*/{ 0x08, OI_RDTSC, OT_NONE, OT_NONE, 0x00010000 },
	/*0x00E5*/{ 0x08, OI_RDMSR, OT_NONE, OT_NONE, 0x00010000 },
	/*0x00E6*/{ 0x08, OI_RDPMC, OT_NONE, OT_NONE, 0x00010000 },
	/*0x00E7*/{ 0x1B, OI_SYSENTER, OT_NONE, OT_NONE, 0x00010000 },
	/*0x00E8*/{ 0x1B, OI_SYSEXIT, OT_NONE, OT_NONE, 0x00010000 },
	/*0x00E9*/{ 0x08, OI_GETSEC, OT_NONE, OT_NONE, 0x00012000 },
	/*0x00EA*/{ 0x1F, OI_CMOVO, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00EB*/{ 0x1F, OI_CMOVNO, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00EC*/{ 0x1F, OI_CMOVB, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00ED*/{ 0x1F, OI_CMOVAE, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00EE*/{ 0x1F, OI_CMOVZ, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00EF*/{ 0x1F, OI_CMOVNZ, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00F0*/{ 0x1F, OI_CMOVBE, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00F1*/{ 0x1F, OI_CMOVA, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00F2*/{ 0x1F, OI_CMOVS, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00F3*/{ 0x1F, OI_CMOVNS, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00F4*/{ 0x1F, OI_CMOVP, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00F5*/{ 0x1F, OI_CMOVNP, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00F6*/{ 0x1F, OI_CMOVL, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00F7*/{ 0x1F, OI_CMOVGE, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00F8*/{ 0x1F, OI_CMOVLE, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00F9*/{ 0x1F, OI_CMOVG, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x00FA*/{ 0x0D, OI_JO, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x00FB*/{ 0x0D, OI_JNO, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x00FC*/{ 0x0D, OI_JB, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x00FD*/{ 0x0D, OI_JNB, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x00FE*/{ 0x0D, OI_JZ, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x00FF*/{ 0x0D, OI_JNZ, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x0100*/{ 0x0D, OI_JBE, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x0101*/{ 0x0D, OI_JNBE, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x0102*/{ 0x0D, OI_JS, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x0103*/{ 0x0D, OI_JNS, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x0104*/{ 0x0D, OI_JP, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x0105*/{ 0x0D, OI_JNP, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x0106*/{ 0x0D, OI_JL, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x0107*/{ 0x0D, OI_JNL, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x0108*/{ 0x0D, OI_JLE, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x0109*/{ 0x0D, OI_JNLE, OT_REL_16_32, OT_NONE, 0x00010000 },
	/*0x010A*/{ 0x08, OI_SETO, OT_RM8, OT_NONE, 0x80012000 },
	/*0x010B*/{ 0x08, OI_SETNO, OT_RM8, OT_NONE, 0x80012000 },
	/*0x010C*/{ 0x08, OI_SETB, OT_RM8, OT_NONE, 0x80012000 },
	/*0x010D*/{ 0x08, OI_SETAE, OT_RM8, OT_NONE, 0x80012000 },
	/*0x010E*/{ 0x08, OI_SETZ, OT_RM8, OT_NONE, 0x80012000 },
	/*0x010F*/{ 0x08, OI_SETNZ, OT_RM8, OT_NONE, 0x80012000 },
	/*0x0110*/{ 0x08, OI_SETBE, OT_RM8, OT_NONE, 0x80012000 },
	/*0x0111*/{ 0x08, OI_SETA, OT_RM8, OT_NONE, 0x80012000 },
	/*0x0112*/{ 0x08, OI_SETS, OT_RM8, OT_NONE, 0x80012000 },
	/*0x0113*/{ 0x08, OI_SETNS, OT_RM8, OT_NONE, 0x80012000 },
	/*0x0114*/{ 0x08, OI_SETP, OT_RM8, OT_NONE, 0x80012000 },
	/*0x0115*/{ 0x08, OI_SETNP, OT_RM8, OT_NONE, 0x80012000 },
	/*0x0116*/{ 0x08, OI_SETL, OT_RM8, OT_NONE, 0x80012000 },
	/*0x0117*/{ 0x08, OI_SETGE, OT_RM8, OT_NONE, 0x80012000 },
	/*0x0118*/{ 0x08, OI_SETLE, OT_RM8, OT_NONE, 0x80012000 },
	/*0x0119*/{ 0x08, OI_SETG, OT_RM8, OT_NONE, 0x80012000 },
	/*0x011A*/{ 0x08, OI_PUSH, OT_SEG_PF, OT_NONE, 0x04010080 },
	/*0x011B*/{ 0x08, OI_POP, OT_SEG_PF, OT_NONE, 0x84010080 },
	/*0x011C*/{ 0x08, OI_CPUID, OT_NONE, OT_NONE, 0x00010000 },
	/*0x011D*/{ 0x08, OI_BT, OT_RM_ALL, OT_REG_ALL, 0x80012000 },
	/*0x011E*/{ 0x08, OI_PUSH, OT_SEG_PF, OT_NONE, 0x04010100 },
	/*0x011F*/{ 0x08, OI_POP, OT_SEG_PF, OT_NONE, 0x84010100 },
	/*0x0120*/{ 0x08, OI_RSM, OT_NONE, OT_NONE, 0x00010000 },
	/*0x0121*/{ 0x08, OI_BTS, OT_RM_ALL, OT_REG_ALL, 0x80012001 },
	/*0x0122*/{ 0x08, OI_IMUL, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x0123*/{ 0x08, OI_CMPXCHG, OT_RM8, OT_REG8, 0x80012001 },
	/*0x0124*/{ 0x08, OI_CMPXCHG, OT_RM_ALL, OT_REG_ALL, 0x80012001 },
	/*0x0125*/{ 0x08, OI_LSS, OT_REG_ALL, OT_MEM16_ALL, 0x04012800 },
	/*0x0126*/{ 0x08, OI_BTR, OT_RM_ALL, OT_REG_ALL, 0x80012001 },
	/*0x0127*/{ 0x08, OI_LFS, OT_REG_ALL, OT_MEM16_ALL, 0x04012800 },
	/*0x0128*/{ 0x08, OI_LGS, OT_REG_ALL, OT_MEM16_ALL, 0x04012800 },
	/*0x0129*/{ 0x08, OI_MOVZX, OT_REG_ALL, OT_RM8, 0x80012000 },
	/*0x012A*/{ 0x08, OI_MOVZX, OT_REG_ALL, OT_RM16, 0x84012800 },
	/*0x012B*/{ 0x0E, OI_UD2, OT_NONE, OT_NONE, 0x00010000 },
	/*0x012C*/{ 0x08, OI_BTC, OT_RM_ALL, OT_REG_ALL, 0x80012001 },
	/*0x012D*/{ 0x08, OI_MOVSX, OT_REG_ALL, OT_RM8, 0x80012000 },
	/*0x012E*/{ 0x08, OI_MOVSX, OT_REG_ALL, OT_RM16, 0x84012800 },
	/*0x012F*/{ 0x08, OI_XADD, OT_RM8, OT_REG8, 0x80012001 },
	/*0x0130*/{ 0x08, OI_XADD, OT_RM_ALL, OT_REG_ALL, 0x80012001 },
	/*0x0131*/{ 0x30, OI_MOVNTI, OT_MEM32_64, OT_REG32_64, 0x84012800 },
	/*0x0132*/{ 0x08, OI_BSWAP, OT_OP_R_ALL, OT_NONE, 0x84010800 },
	/*0x0133*/{ 0x08, OI_BSWAP, OT_OP_R_ALL, OT_NONE, 0x84010800 },
	/*0x0134*/{ 0x08, OI_BSWAP, OT_OP_R_ALL, OT_NONE, 0x84010800 },
	/*0x0135*/{ 0x08, OI_BSWAP, OT_OP_R_ALL, OT_NONE, 0x84010800 },
	/*0x0136*/{ 0x08, OI_BSWAP, OT_OP_R_ALL, OT_NONE, 0x84010800 },
	/*0x0137*/{ 0x08, OI_BSWAP, OT_OP_R_ALL, OT_NONE, 0x84010800 },
	/*0x0138*/{ 0x08, OI_BSWAP, OT_OP_R_ALL, OT_NONE, 0x84010800 },
	/*0x0139*/{ 0x08, OI_BSWAP, OT_OP_R_ALL, OT_NONE, 0x84010800 },
	/*0x013A*/{ 0x08, OI_ADD, OT_RM8, OT_IMM8, 0xC0002001 },
	/*0x013B*/{ 0x08, OI_OR, OT_RM8, OT_IMM8, 0xC0002001 },
	/*0x013C*/{ 0x08, OI_ADC, OT_RM8, OT_IMM8, 0xC0002001 },
	/*0x013D*/{ 0x08, OI_SBB, OT_RM8, OT_IMM8, 0xC0002001 },
	/*0x013E*/{ 0x08, OI_AND, OT_RM8, OT_IMM8, 0xC0002001 },
	/*0x013F*/{ 0x08, OI_SUB, OT_RM8, OT_IMM8, 0xC0002001 },
	/*0x0140*/{ 0x08, OI_XOR, OT_RM8, OT_IMM8, 0xC0002001 },
	/*0x0141*/{ 0x08, OI_CMP, OT_RM8, OT_IMM8, 0x40002000 },
	/*0x0142*/{ 0x08, OI_ADD, OT_RM_ALL, OT_IMM_ALL, 0xC0002001 },
	/*0x0143*/{ 0x08, OI_OR, OT_RM_ALL, OT_IMM_ALL, 0xC0002001 },
	/*0x0144*/{ 0x08, OI_ADC, OT_RM_ALL, OT_IMM_ALL, 0xC0002001 },
	/*0x0145*/{ 0x08, OI_SBB, OT_RM_ALL, OT_IMM_ALL, 0xC0002001 },
	/*0x0146*/{ 0x08, OI_AND, OT_RM_ALL, OT_IMM_ALL, 0xC0002001 },
	/*0x0147*/{ 0x08, OI_SUB, OT_RM_ALL, OT_IMM_ALL, 0xC0002001 },
	/*0x0148*/{ 0x08, OI_XOR, OT_RM_ALL, OT_IMM_ALL, 0xC0002001 },
	/*0x0149*/{ 0x08, OI_CMP, OT_RM_ALL, OT_IMM_ALL, 0x40002000 },
	/*0x014A*/{ 0x08, OI_ADD, OT_RM8, OT_IMM8, 0xC2002001 },
	/*0x014B*/{ 0x08, OI_OR, OT_RM8, OT_IMM8, 0xC2002001 },
	/*0x014C*/{ 0x08, OI_ADC, OT_RM8, OT_IMM8, 0xC2002001 },
	/*0x014D*/{ 0x08, OI_SBB, OT_RM8, OT_IMM8, 0xC2002001 },
	/*0x014E*/{ 0x08, OI_AND, OT_RM8, OT_IMM8, 0xC2002001 },
	/*0x014F*/{ 0x08, OI_SUB, OT_RM8, OT_IMM8, 0xC2002001 },
	/*0x0150*/{ 0x08, OI_XOR, OT_RM8, OT_IMM8, 0xC2002001 },
	/*0x0151*/{ 0x08, OI_CMP, OT_RM8, OT_IMM8, 0x42002000 },
	/*0x0152*/{ 0x08, OI_ADD, OT_RM_ALL, OT_SE_IMM8, 0xC0002001 },
	/*0x0153*/{ 0x08, OI_OR, OT_RM_ALL, OT_SE_IMM8, 0xC0012001 },
	/*0x0154*/{ 0x08, OI_ADC, OT_RM_ALL, OT_SE_IMM8, 0xC0002001 },
	/*0x0155*/{ 0x08, OI_SBB, OT_RM_ALL, OT_SE_IMM8, 0xC0002001 },
	/*0x0156*/{ 0x08, OI_AND, OT_RM_ALL, OT_SE_IMM8, 0xC0012001 },
	/*0x0157*/{ 0x08, OI_SUB, OT_RM_ALL, OT_SE_IMM8, 0xC0002001 },
	/*0x0158*/{ 0x08, OI_XOR, OT_RM_ALL, OT_SE_IMM8, 0xC0012001 },
	/*0x0159*/{ 0x08, OI_CMP, OT_RM_ALL, OT_SE_IMM8, 0x40002000 },
	/*0x015A*/{ 0x08, OI_POP, OT_RM_ALL, OT_NONE, 0xC4002000 },
	/*0x015B*/{ 0x08, OI_ROL, OT_RM8, OT_IMM8, 0xC0002000 },
	/*0x015C*/{ 0x08, OI_ROR, OT_RM8, OT_IMM8, 0xC0002000 },
	/*0x015D*/{ 0x08, OI_RCL, OT_RM8, OT_IMM8, 0xC0002000 },
	/*0x015E*/{ 0x08, OI_RCR, OT_RM8, OT_IMM8, 0xC0002000 },
	/*0x015F*/{ 0x08, OI_SHL, OT_RM8, OT_IMM8, 0xC0002000 },
	/*0x0160*/{ 0x08, OI_SHR, OT_RM8, OT_IMM8, 0xC0002000 },
	/*0x0161*/{ 0x08, OI_SAL, OT_RM8, OT_IMM8, 0xC0002000 },
	/*0x0162*/{ 0x08, OI_SAR, OT_RM8, OT_IMM8, 0xC0002000 },
	/*0x0163*/{ 0x08, OI_ROL, OT_RM_ALL, OT_IMM8, 0xC0002000 },
	/*0x0164*/{ 0x08, OI_ROR, OT_RM_ALL, OT_IMM8, 0xC0002000 },
	/*0x0165*/{ 0x08, OI_RCL, OT_RM_ALL, OT_IMM8, 0xC0002000 },
	/*0x0166*/{ 0x08, OI_RCR, OT_RM_ALL, OT_IMM8, 0xC0002000 },
	/*0x0167*/{ 0x08, OI_SHL, OT_RM_ALL, OT_IMM8, 0xC0002000 },
	/*0x0168*/{ 0x08, OI_SHR, OT_RM_ALL, OT_IMM8, 0xC0002000 },
	/*0x0169*/{ 0x08, OI_SAL, OT_RM_ALL, OT_IMM8, 0xC0002000 },
	/*0x016A*/{ 0x08, OI_SAR, OT_RM_ALL, OT_IMM8, 0xC0002000 },
	/*0x016B*/{ 0x08, OI_MOV, OT_RM8, OT_IMM8, 0xC0002000 },
	/*0x016C*/{ 0x08, OI_MOV, OT_RM_ALL, OT_IMM_ALL, 0xC0002000 },
	/*0x016D*/{ 0x08, OI_ROL, OT_RM8, OT_CONST1, 0xC0002000 },
	/*0x016E*/{ 0x08, OI_ROR, OT_RM8, OT_CONST1, 0xC0002000 },
	/*0x016F*/{ 0x08, OI_RCL, OT_RM8, OT_CONST1, 0xC0002000 },
	/*0x0170*/{ 0x08, OI_RCR, OT_RM8, OT_CONST1, 0xC0002000 },
	/*0x0171*/{ 0x08, OI_SHL, OT_RM8, OT_CONST1, 0xC0002000 },
	/*0x0172*/{ 0x08, OI_SHR, OT_RM8, OT_CONST1, 0xC0002000 },
	/*0x0173*/{ 0x08, OI_SAL, OT_RM8, OT_CONST1, 0xC0002000 },
	/*0x0174*/{ 0x08, OI_SAR, OT_RM8, OT_CONST1, 0xC0002000 },
	/*0x0175*/{ 0x08, OI_ROL, OT_RM_ALL, OT_CONST1, 0xC0002000 },
	/*0x0176*/{ 0x08, OI_ROR, OT_RM_ALL, OT_CONST1, 0xC0002000 },
	/*0x0177*/{ 0x08, OI_RCL, OT_RM_ALL, OT_CONST1, 0xC0002000 },
	/*0x0178*/{ 0x08, OI_RCR, OT_RM_ALL, OT_CONST1, 0xC0002000 },
	/*0x0179*/{ 0x08, OI_SHL, OT_RM_ALL, OT_CONST1, 0xC0002000 },
	/*0x017A*/{ 0x08, OI_SHR, OT_RM_ALL, OT_CONST1, 0xC0002000 },
	/*0x017B*/{ 0x08, OI_SAL, OT_RM_ALL, OT_CONST1, 0xC0002000 },
	/*0x017C*/{ 0x08, OI_SAR, OT_RM_ALL, OT_CONST1, 0xC0002000 },
	/*0x017D*/{ 0x08, OI_ROL, OT_RM8, OT_REG_CL, 0xC0002000 },
	/*0x017E*/{ 0x08, OI_ROR, OT_RM8, OT_REG_CL, 0xC0002000 },
	/*0x017F*/{ 0x08, OI_RCL, OT_RM8, OT_REG_CL, 0xC0002000 },
	/*0x0180*/{ 0x08, OI_RCR, OT_RM8, OT_REG_CL, 0xC0002000 },
	/*0x0181*/{ 0x08, OI_SHL, OT_RM8, OT_REG_CL, 0xC0002000 },
	/*0x0182*/{ 0x08, OI_SHR, OT_RM8, OT_REG_CL, 0xC0002000 },
	/*0x0183*/{ 0x08, OI_SAL, OT_RM8, OT_REG_CL, 0xC0002000 },
	/*0x0184*/{ 0x08, OI_SAR, OT_RM8, OT_REG_CL, 0xC0002000 },
	/*0x0185*/{ 0x08, OI_ROL, OT_RM_ALL, OT_REG_CL, 0xC0002000 },
	/*0x0186*/{ 0x08, OI_ROR, OT_RM_ALL, OT_REG_CL, 0xC0002000 },
	/*0x0187*/{ 0x08, OI_RCL, OT_RM_ALL, OT_REG_CL, 0xC0002000 },
	/*0x0188*/{ 0x08, OI_RCR, OT_RM_ALL, OT_REG_CL, 0xC0002000 },
	/*0x0189*/{ 0x08, OI_SHL, OT_RM_ALL, OT_REG_CL, 0xC0002000 },
	/*0x018A*/{ 0x08, OI_SHR, OT_RM_ALL, OT_REG_CL, 0xC0002000 },
	/*0x018B*/{ 0x08, OI_SAL, OT_RM_ALL, OT_REG_CL, 0xC0002000 },
	/*0x018C*/{ 0x08, OI_SAR, OT_RM_ALL, OT_REG_CL, 0xC0002000 },
	/*0x018D*/{ 0x10, OI_FADD, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x018E*/{ 0x10, OI_FMUL, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x018F*/{ 0x10, OI_FCOM, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x0190*/{ 0x10, OI_FCOMP, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x0191*/{ 0x10, OI_FSUB, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x0192*/{ 0x10, OI_FSUBR, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x0193*/{ 0x10, OI_FDIV, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x0194*/{ 0x10, OI_FDIVR, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x0195*/{ 0x10, OI_FADD, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x0196*/{ 0x10, OI_FADD, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x0197*/{ 0x10, OI_FADD, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x0198*/{ 0x10, OI_FADD, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x0199*/{ 0x10, OI_FADD, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x019A*/{ 0x10, OI_FADD, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x019B*/{ 0x10, OI_FADD, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x019C*/{ 0x10, OI_FADD, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x019D*/{ 0x10, OI_FMUL, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x019E*/{ 0x10, OI_FMUL, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x019F*/{ 0x10, OI_FMUL, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01A0*/{ 0x10, OI_FMUL, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01A1*/{ 0x10, OI_FMUL, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01A2*/{ 0x10, OI_FMUL, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01A3*/{ 0x10, OI_FMUL, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01A4*/{ 0x10, OI_FMUL, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01A5*/{ 0x10, OI_FCOM, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01A6*/{ 0x10, OI_FCOM, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01A7*/{ 0x10, OI_FCOM, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01A8*/{ 0x10, OI_FCOM, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01A9*/{ 0x10, OI_FCOM, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01AA*/{ 0x10, OI_FCOM, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01AB*/{ 0x10, OI_FCOM, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01AC*/{ 0x10, OI_FCOM, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01AD*/{ 0x10, OI_FCOMP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01AE*/{ 0x10, OI_FCOMP, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01AF*/{ 0x10, OI_FCOMP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01B0*/{ 0x10, OI_FCOMP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01B1*/{ 0x10, OI_FCOMP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01B2*/{ 0x10, OI_FCOMP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01B3*/{ 0x10, OI_FCOMP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01B4*/{ 0x10, OI_FCOMP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01B5*/{ 0x10, OI_FSUB, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01B6*/{ 0x10, OI_FSUB, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01B7*/{ 0x10, OI_FSUB, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01B8*/{ 0x10, OI_FSUB, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01B9*/{ 0x10, OI_FSUB, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01BA*/{ 0x10, OI_FSUB, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01BB*/{ 0x10, OI_FSUB, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01BC*/{ 0x10, OI_FSUB, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01BD*/{ 0x10, OI_FSUBR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01BE*/{ 0x10, OI_FSUBR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01BF*/{ 0x10, OI_FSUBR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01C0*/{ 0x10, OI_FSUBR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01C1*/{ 0x10, OI_FSUBR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01C2*/{ 0x10, OI_FSUBR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01C3*/{ 0x10, OI_FSUBR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01C4*/{ 0x10, OI_FSUBR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01C5*/{ 0x10, OI_FDIV, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01C6*/{ 0x10, OI_FDIV, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01C7*/{ 0x10, OI_FDIV, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01C8*/{ 0x10, OI_FDIV, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01C9*/{ 0x10, OI_FDIV, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01CA*/{ 0x10, OI_FDIV, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01CB*/{ 0x10, OI_FDIV, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01CC*/{ 0x10, OI_FDIV, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01CD*/{ 0x10, OI_FDIVR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01CE*/{ 0x10, OI_FDIVR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01CF*/{ 0x10, OI_FDIVR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01D0*/{ 0x10, OI_FDIVR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01D1*/{ 0x10, OI_FDIVR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01D2*/{ 0x10, OI_FDIVR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01D3*/{ 0x10, OI_FDIVR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01D4*/{ 0x10, OI_FDIVR, OT_FREG_SI0, OT_FREG_SI, 0x40000000 },
	/*0x01D5*/{ 0x10, OI_FLD, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x01D6*/{ 0x10, OI_FST, OT_FPUM32, OT_NONE, 0xC0002000 },
	/*0x01D7*/{ 0x10, OI_FSTP, OT_FPUM32, OT_NONE, 0xC0002000 },
	/*0x01D8*/{ 0x10, OI_FLDENV, OT_MEM, OT_NONE, 0x40002000 },
	/*0x01D9*/{ 0x10, OI_FLDCW, OT_FPUM16, OT_NONE, 0x40002000 },
	/*0x01DA*/{ 0x10, OI_FLD, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01DB*/{ 0x10, OI_FLD, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01DC*/{ 0x10, OI_FLD, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01DD*/{ 0x10, OI_FLD, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01DE*/{ 0x10, OI_FLD, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01DF*/{ 0x10, OI_FLD, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01E0*/{ 0x10, OI_FLD, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01E1*/{ 0x10, OI_FLD, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01E2*/{ 0x10, OI_FXCH, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01E3*/{ 0x10, OI_FXCH, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01E4*/{ 0x10, OI_FXCH, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01E5*/{ 0x10, OI_FXCH, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01E6*/{ 0x10, OI_FXCH, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01E7*/{ 0x10, OI_FXCH, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01E8*/{ 0x10, OI_FXCH, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01E9*/{ 0x10, OI_FXCH, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x01EA*/{ 0x10, OI_FNOP, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01EB*/{ 0x10, OI_FCHS, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01EC*/{ 0x10, OI_FABS, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01ED*/{ 0x10, OI_FTST, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01EE*/{ 0x10, OI_FXAM, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01EF*/{ 0x10, OI_FLD1, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01F0*/{ 0x10, OI_FLDL2T, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01F1*/{ 0x10, OI_FLDL2E, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01F2*/{ 0x10, OI_FLDPI, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01F3*/{ 0x10, OI_FLDLG2, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01F4*/{ 0x10, OI_FLDLN2, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01F5*/{ 0x10, OI_FLDZ, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01F6*/{ 0x10, OI_F2XM1, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01F7*/{ 0x10, OI_FYL2X, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01F8*/{ 0x10, OI_FPTAN, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01F9*/{ 0x10, OI_FPATAN, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01FA*/{ 0x10, OI_FXTRACT, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01FB*/{ 0x10, OI_FPREM1, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01FC*/{ 0x10, OI_FDECSTP, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01FD*/{ 0x10, OI_FINCSTP, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01FE*/{ 0x10, OI_FPREM, OT_NONE, OT_NONE, 0x40000000 },
	/*0x01FF*/{ 0x10, OI_FYL2XP1, OT_NONE, OT_NONE, 0x40000000 },
	/*0x0200*/{ 0x10, OI_FSQRT, OT_NONE, OT_NONE, 0x40000000 },
	/*0x0201*/{ 0x10, OI_FSINCOS, OT_NONE, OT_NONE, 0x40000000 },
	/*0x0202*/{ 0x10, OI_FRNDINT, OT_NONE, OT_NONE, 0x40000000 },
	/*0x0203*/{ 0x10, OI_FSCALE, OT_NONE, OT_NONE, 0x40000000 },
	/*0x0204*/{ 0x10, OI_FSIN, OT_NONE, OT_NONE, 0x40000000 },
	/*0x0205*/{ 0x10, OI_FCOS, OT_NONE, OT_NONE, 0x40000000 },
	/*0x0206*/{ 0x10, OI_FIADD, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x0207*/{ 0x10, OI_FIMUL, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x0208*/{ 0x10, OI_FICOM, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x0209*/{ 0x10, OI_FICOMP, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x020A*/{ 0x10, OI_FISUB, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x020B*/{ 0x10, OI_FISUBR, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x020C*/{ 0x10, OI_FIDIV, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x020D*/{ 0x10, OI_FIDIVR, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x020E*/{ 0x18, OI_FCMOVB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x020F*/{ 0x18, OI_FCMOVB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0210*/{ 0x18, OI_FCMOVB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0211*/{ 0x18, OI_FCMOVB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0212*/{ 0x18, OI_FCMOVB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0213*/{ 0x18, OI_FCMOVB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0214*/{ 0x18, OI_FCMOVB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0215*/{ 0x18, OI_FCMOVB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0216*/{ 0x18, OI_FCMOVE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0217*/{ 0x18, OI_FCMOVE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0218*/{ 0x18, OI_FCMOVE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0219*/{ 0x18, OI_FCMOVE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x021A*/{ 0x18, OI_FCMOVE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x021B*/{ 0x18, OI_FCMOVE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x021C*/{ 0x18, OI_FCMOVE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x021D*/{ 0x18, OI_FCMOVE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x021E*/{ 0x18, OI_FCMOVBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x021F*/{ 0x18, OI_FCMOVBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0220*/{ 0x18, OI_FCMOVBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0221*/{ 0x18, OI_FCMOVBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0222*/{ 0x18, OI_FCMOVBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0223*/{ 0x18, OI_FCMOVBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0224*/{ 0x18, OI_FCMOVBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0225*/{ 0x18, OI_FCMOVBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0226*/{ 0x18, OI_FCMOVU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0227*/{ 0x18, OI_FCMOVU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0228*/{ 0x18, OI_FCMOVU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0229*/{ 0x18, OI_FCMOVU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x022A*/{ 0x18, OI_FCMOVU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x022B*/{ 0x18, OI_FCMOVU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x022C*/{ 0x18, OI_FCMOVU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x022D*/{ 0x18, OI_FCMOVU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x022E*/{ 0x10, OI_FUCOMPP, OT_NONE, OT_NONE, 0x40000000 },
	/*0x022F*/{ 0x10, OI_FILD, OT_FPUM32, OT_NONE, 0x40002000 },
	/*0x0230*/{ 0x38, OI_FISTTP, OT_FPUM32, OT_NONE, 0x40012000 },
	/*0x0231*/{ 0x10, OI_FIST, OT_FPUM32, OT_NONE, 0xC0002000 },
	/*0x0232*/{ 0x10, OI_FISTP, OT_FPUM32, OT_NONE, 0xC0002000 },
	/*0x0233*/{ 0x10, OI_FLD, OT_FPUM80, OT_NONE, 0x40002000 },
	/*0x0234*/{ 0x10, OI_FSTP, OT_FPUM80, OT_NONE, 0xC0002000 },
	/*0x0235*/{ 0x18, OI_FCMOVNB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0236*/{ 0x18, OI_FCMOVNB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0237*/{ 0x18, OI_FCMOVNB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0238*/{ 0x18, OI_FCMOVNB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0239*/{ 0x18, OI_FCMOVNB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x023A*/{ 0x18, OI_FCMOVNB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x023B*/{ 0x18, OI_FCMOVNB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x023C*/{ 0x18, OI_FCMOVNB, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x023D*/{ 0x18, OI_FCMOVNE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x023E*/{ 0x18, OI_FCMOVNE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x023F*/{ 0x18, OI_FCMOVNE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0240*/{ 0x18, OI_FCMOVNE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0241*/{ 0x18, OI_FCMOVNE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0242*/{ 0x18, OI_FCMOVNE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0243*/{ 0x18, OI_FCMOVNE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0244*/{ 0x18, OI_FCMOVNE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0245*/{ 0x18, OI_FCMOVNBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0246*/{ 0x18, OI_FCMOVNBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0247*/{ 0x18, OI_FCMOVNBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0248*/{ 0x18, OI_FCMOVNBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0249*/{ 0x18, OI_FCMOVNBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x024A*/{ 0x18, OI_FCMOVNBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x024B*/{ 0x18, OI_FCMOVNBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x024C*/{ 0x18, OI_FCMOVNBE, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x024D*/{ 0x18, OI_FCMOVNU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x024E*/{ 0x18, OI_FCMOVNU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x024F*/{ 0x18, OI_FCMOVNU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0250*/{ 0x18, OI_FCMOVNU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0251*/{ 0x18, OI_FCMOVNU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0252*/{ 0x18, OI_FCMOVNU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0253*/{ 0x18, OI_FCMOVNU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0254*/{ 0x18, OI_FCMOVNU, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0255*/{ 0x10, OI_FENI, OT_NONE, OT_NONE, 0x40000000 },
	/*0x0256*/{ 0x10, OI_FEDISI, OT_NONE, OT_NONE, 0x40000000 },
	/*0x0257*/{ 0x10, OI_FSETPM, OT_NONE, OT_NONE, 0x40000000 },
	/*0x0258*/{ 0x10, OI_FUCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0259*/{ 0x10, OI_FUCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x025A*/{ 0x10, OI_FUCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x025B*/{ 0x10, OI_FUCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x025C*/{ 0x10, OI_FUCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x025D*/{ 0x10, OI_FUCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x025E*/{ 0x10, OI_FUCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x025F*/{ 0x10, OI_FUCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0260*/{ 0x18, OI_FCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0261*/{ 0x18, OI_FCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0262*/{ 0x18, OI_FCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0263*/{ 0x18, OI_FCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0264*/{ 0x18, OI_FCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0265*/{ 0x18, OI_FCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0266*/{ 0x18, OI_FCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0267*/{ 0x18, OI_FCOMI, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0268*/{ 0x10, OI_FADD, OT_FPUM64, OT_NONE, 0x40002000 },
	/*0x0269*/{ 0x10, OI_FMUL, OT_FPUM64, OT_NONE, 0x40002000 },
	/*0x026A*/{ 0x10, OI_FCOM, OT_FPUM64, OT_NONE, 0x40002000 },
	/*0x026B*/{ 0x10, OI_FCOMP, OT_FPUM64, OT_NONE, 0x40002000 },
	/*0x026C*/{ 0x10, OI_FSUB, OT_FPUM64, OT_NONE, 0x40002000 },
	/*0x026D*/{ 0x10, OI_FSUBR, OT_FPUM64, OT_NONE, 0x40002000 },
	/*0x026E*/{ 0x10, OI_FDIV, OT_FPUM64, OT_NONE, 0x40002000 },
	/*0x026F*/{ 0x10, OI_FDIVR, OT_FPUM64, OT_NONE, 0x40002000 },
	/*0x0270*/{ 0x10, OI_FADD, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0271*/{ 0x10, OI_FADD, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0272*/{ 0x10, OI_FADD, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0273*/{ 0x10, OI_FADD, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0274*/{ 0x10, OI_FADD, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0275*/{ 0x10, OI_FADD, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0276*/{ 0x10, OI_FADD, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0277*/{ 0x10, OI_FADD, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0278*/{ 0x10, OI_FMUL, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0279*/{ 0x10, OI_FMUL, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x027A*/{ 0x10, OI_FMUL, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x027B*/{ 0x10, OI_FMUL, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x027C*/{ 0x10, OI_FMUL, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x027D*/{ 0x10, OI_FMUL, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x027E*/{ 0x10, OI_FMUL, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x027F*/{ 0x10, OI_FMUL, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0280*/{ 0x10, OI_FSUBR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0281*/{ 0x10, OI_FSUBR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0282*/{ 0x10, OI_FSUBR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0283*/{ 0x10, OI_FSUBR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0284*/{ 0x10, OI_FSUBR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0285*/{ 0x10, OI_FSUBR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0286*/{ 0x10, OI_FSUBR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0287*/{ 0x10, OI_FSUBR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0288*/{ 0x10, OI_FSUB, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0289*/{ 0x10, OI_FSUB, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x028A*/{ 0x10, OI_FSUB, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x028B*/{ 0x10, OI_FSUB, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x028C*/{ 0x10, OI_FSUB, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x028D*/{ 0x10, OI_FSUB, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x028E*/{ 0x10, OI_FSUB, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x028F*/{ 0x10, OI_FSUB, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0290*/{ 0x10, OI_FDIVR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0291*/{ 0x10, OI_FDIVR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0292*/{ 0x10, OI_FDIVR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0293*/{ 0x10, OI_FDIVR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0294*/{ 0x10, OI_FDIVR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0295*/{ 0x10, OI_FDIVR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0296*/{ 0x10, OI_FDIVR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0297*/{ 0x10, OI_FDIVR, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0298*/{ 0x10, OI_FDIV, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0299*/{ 0x10, OI_FDIV, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x029A*/{ 0x10, OI_FDIV, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x029B*/{ 0x10, OI_FDIV, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x029C*/{ 0x10, OI_FDIV, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x029D*/{ 0x10, OI_FDIV, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x029E*/{ 0x10, OI_FDIV, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x029F*/{ 0x10, OI_FDIV, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02A0*/{ 0x10, OI_FLD, OT_FPUM64, OT_NONE, 0x40002000 },
	/*0x02A1*/{ 0x38, OI_FISTTP, OT_FPUM64, OT_NONE, 0x40012000 },
	/*0x02A2*/{ 0x10, OI_FST, OT_FPUM64, OT_NONE, 0xC0002000 },
	/*0x02A3*/{ 0x10, OI_FSTP, OT_FPUM64, OT_NONE, 0xC0002000 },
	/*0x02A4*/{ 0x10, OI_FRSTOR, OT_MEM, OT_NONE, 0x40002000 },
	/*0x02A5*/{ 0x10, OI_FFREE, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02A6*/{ 0x10, OI_FFREE, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02A7*/{ 0x10, OI_FFREE, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02A8*/{ 0x10, OI_FFREE, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02A9*/{ 0x10, OI_FFREE, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02AA*/{ 0x10, OI_FFREE, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02AB*/{ 0x10, OI_FFREE, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02AC*/{ 0x10, OI_FFREE, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02AD*/{ 0x10, OI_FST, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02AE*/{ 0x10, OI_FST, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02AF*/{ 0x10, OI_FST, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02B0*/{ 0x10, OI_FST, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02B1*/{ 0x10, OI_FST, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02B2*/{ 0x10, OI_FST, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02B3*/{ 0x10, OI_FST, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02B4*/{ 0x10, OI_FST, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02B5*/{ 0x10, OI_FSTP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02B6*/{ 0x10, OI_FSTP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02B7*/{ 0x10, OI_FSTP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02B8*/{ 0x10, OI_FSTP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02B9*/{ 0x10, OI_FSTP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02BA*/{ 0x10, OI_FSTP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02BB*/{ 0x10, OI_FSTP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02BC*/{ 0x10, OI_FSTP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02BD*/{ 0x10, OI_FUCOM, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02BE*/{ 0x10, OI_FUCOM, OT_NONE, OT_NONE, 0x40000000 },
	/*0x02BF*/{ 0x10, OI_FUCOM, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02C0*/{ 0x10, OI_FUCOM, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02C1*/{ 0x10, OI_FUCOM, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02C2*/{ 0x10, OI_FUCOM, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02C3*/{ 0x10, OI_FUCOM, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02C4*/{ 0x10, OI_FUCOM, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02C5*/{ 0x10, OI_FUCOMP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02C6*/{ 0x10, OI_FUCOMP, OT_NONE, OT_NONE, 0x40000000 },
	/*0x02C7*/{ 0x10, OI_FUCOMP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02C8*/{ 0x10, OI_FUCOMP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02C9*/{ 0x10, OI_FUCOMP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02CA*/{ 0x10, OI_FUCOMP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02CB*/{ 0x10, OI_FUCOMP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02CC*/{ 0x10, OI_FUCOMP, OT_FREG_SI, OT_NONE, 0x40000000 },
	/*0x02CD*/{ 0x10, OI_FIADD, OT_FPUM16, OT_NONE, 0x40002000 },
	/*0x02CE*/{ 0x10, OI_FIMUL, OT_FPUM16, OT_NONE, 0x40002000 },
	/*0x02CF*/{ 0x10, OI_FICOM, OT_FPUM16, OT_NONE, 0x40002000 },
	/*0x02D0*/{ 0x10, OI_FICOMP, OT_FPUM16, OT_NONE, 0x40002000 },
	/*0x02D1*/{ 0x10, OI_FISUB, OT_FPUM16, OT_NONE, 0x40002000 },
	/*0x02D2*/{ 0x10, OI_FISUBR, OT_FPUM16, OT_NONE, 0x40002000 },
	/*0x02D3*/{ 0x10, OI_FIDIV, OT_FPUM16, OT_NONE, 0x40002000 },
	/*0x02D4*/{ 0x10, OI_FIDIVR, OT_FPUM16, OT_NONE, 0x40002000 },
	/*0x02D5*/{ 0x10, OI_FADDP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02D6*/{ 0x10, OI_FADDP, OT_NONE, OT_NONE, 0x40000000 },
	/*0x02D7*/{ 0x10, OI_FADDP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02D8*/{ 0x10, OI_FADDP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02D9*/{ 0x10, OI_FADDP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02DA*/{ 0x10, OI_FADDP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02DB*/{ 0x10, OI_FADDP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02DC*/{ 0x10, OI_FADDP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02DD*/{ 0x10, OI_FMULP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02DE*/{ 0x10, OI_FMULP, OT_NONE, OT_NONE, 0x40000000 },
	/*0x02DF*/{ 0x10, OI_FMULP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02E0*/{ 0x10, OI_FMULP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02E1*/{ 0x10, OI_FMULP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02E2*/{ 0x10, OI_FMULP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02E3*/{ 0x10, OI_FMULP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02E4*/{ 0x10, OI_FMULP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02E5*/{ 0x10, OI_FCOMPP, OT_NONE, OT_NONE, 0x40000000 },
	/*0x02E6*/{ 0x10, OI_FSUBRP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02E7*/{ 0x10, OI_FSUBRP, OT_NONE, OT_NONE, 0x40000000 },
	/*0x02E8*/{ 0x10, OI_FSUBRP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02E9*/{ 0x10, OI_FSUBRP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02EA*/{ 0x10, OI_FSUBRP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02EB*/{ 0x10, OI_FSUBRP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02EC*/{ 0x10, OI_FSUBRP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02ED*/{ 0x10, OI_FSUBRP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02EE*/{ 0x10, OI_FSUBP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02EF*/{ 0x10, OI_FSUBP, OT_NONE, OT_NONE, 0x40000000 },
	/*0x02F0*/{ 0x10, OI_FSUBP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02F1*/{ 0x10, OI_FSUBP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02F2*/{ 0x10, OI_FSUBP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02F3*/{ 0x10, OI_FSUBP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02F4*/{ 0x10, OI_FSUBP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02F5*/{ 0x10, OI_FSUBP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02F6*/{ 0x10, OI_FDIVRP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02F7*/{ 0x10, OI_FDIVRP, OT_NONE, OT_NONE, 0x40000000 },
	/*0x02F8*/{ 0x10, OI_FDIVRP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02F9*/{ 0x10, OI_FDIVRP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02FA*/{ 0x10, OI_FDIVRP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02FB*/{ 0x10, OI_FDIVRP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02FC*/{ 0x10, OI_FDIVRP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02FD*/{ 0x10, OI_FDIVRP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02FE*/{ 0x10, OI_FDIVP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x02FF*/{ 0x10, OI_FDIVP, OT_NONE, OT_NONE, 0x40000000 },
	/*0x0300*/{ 0x10, OI_FDIVP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0301*/{ 0x10, OI_FDIVP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0302*/{ 0x10, OI_FDIVP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0303*/{ 0x10, OI_FDIVP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0304*/{ 0x10, OI_FDIVP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0305*/{ 0x10, OI_FDIVP, OT_FREG_SI, OT_FREG_SI0, 0x40000000 },
	/*0x0306*/{ 0x10, OI_FILD, OT_FPUM16, OT_NONE, 0x40002000 },
	/*0x0307*/{ 0x38, OI_FISTTP, OT_FPUM16, OT_NONE, 0x40012000 },
	/*0x0308*/{ 0x10, OI_FIST, OT_FPUM16, OT_NONE, 0xC0002000 },
	/*0x0309*/{ 0x10, OI_FISTP, OT_FPUM16, OT_NONE, 0xC0002000 },
	/*0x030A*/{ 0x10, OI_FBLD, OT_FPUM80, OT_NONE, 0x40002000 },
	/*0x030B*/{ 0x10, OI_FILD, OT_FPUM64, OT_NONE, 0x40002000 },
	/*0x030C*/{ 0x10, OI_FBSTP, OT_FPUM80, OT_NONE, 0xC0002000 },
	/*0x030D*/{ 0x10, OI_FISTP, OT_FPUM64, OT_NONE, 0xC0002000 },
	/*0x030E*/{ 0x10, OI_FUCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x030F*/{ 0x10, OI_FUCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0310*/{ 0x10, OI_FUCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0311*/{ 0x10, OI_FUCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0312*/{ 0x10, OI_FUCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0313*/{ 0x10, OI_FUCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0314*/{ 0x10, OI_FUCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0315*/{ 0x10, OI_FUCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0316*/{ 0x10, OI_FCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0317*/{ 0x10, OI_FCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0318*/{ 0x10, OI_FCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x0319*/{ 0x10, OI_FCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x031A*/{ 0x10, OI_FCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x031B*/{ 0x10, OI_FCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x031C*/{ 0x10, OI_FCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x031D*/{ 0x10, OI_FCOMIP, OT_FREG_SI0, OT_FREG_SI, 0x40010000 },
	/*0x031E*/{ 0x08, OI_TEST, OT_RM8, OT_IMM8, 0x40002000 },
	/*0x031F*/{ 0x08, OI_NOT, OT_RM8, OT_NONE, 0xC0002001 },
	/*0x0320*/{ 0x08, OI_NEG, OT_RM8, OT_NONE, 0xC0002001 },
	/*0x0321*/{ 0x08, OI_MUL, OT_RM8, OT_NONE, 0xC0002000 },
	/*0x0322*/{ 0x08, OI_IMUL, OT_RM8, OT_NONE, 0xC0002000 },
	/*0x0323*/{ 0x08, OI_DIV, OT_RM8, OT_NONE, 0xC0002000 },
	/*0x0324*/{ 0x08, OI_IDIV, OT_RM8, OT_NONE, 0xC0002000 },
	/*0x0325*/{ 0x08, OI_TEST, OT_RM_ALL, OT_IMM_ALL, 0x40002000 },
	/*0x0326*/{ 0x08, OI_NOT, OT_RM_ALL, OT_NONE, 0xC0002001 },
	/*0x0327*/{ 0x08, OI_NEG, OT_RM_ALL, OT_NONE, 0xC0002001 },
	/*0x0328*/{ 0x08, OI_MUL, OT_RM_ALL, OT_NONE, 0xC0002000 },
	/*0x0329*/{ 0x08, OI_IMUL, OT_RM_ALL, OT_NONE, 0xC0002000 },
	/*0x032A*/{ 0x08, OI_DIV, OT_RM_ALL, OT_NONE, 0xC0002000 },
	/*0x032B*/{ 0x08, OI_IDIV, OT_RM_ALL, OT_NONE, 0xC0002000 },
	/*0x032C*/{ 0x08, OI_INC, OT_RM8, OT_NONE, 0xC0002001 },
	/*0x032D*/{ 0x08, OI_DEC, OT_RM8, OT_NONE, 0xC0002001 },
	/*0x032E*/{ 0x08, OI_INC, OT_RM_ALL, OT_NONE, 0xC0002001 },
	/*0x032F*/{ 0x08, OI_DEC, OT_RM_ALL, OT_NONE, 0xC0002001 },
	/*0x0330*/{ 0x09, OI_CALL, OT_RM_ALL, OT_NONE, 0x44002000 },
	/*0x0331*/{ 0x09, OI_CALL_FAR, OT_MEM16_ALL, OT_NONE, 0x44002800 },
	/*0x0332*/{ 0x0C, OI_JMP, OT_RM_ALL, OT_NONE, 0x44002000 },
	/*0x0333*/{ 0x0C, OI_JMP_FAR, OT_MEM16_ALL, OT_NONE, 0x44002800 },
	/*0x0334*/{ 0x08, OI_PUSH, OT_RM_ALL, OT_NONE, 0x44002000 },
	/*0x0335*/{ 0x08, OI_SLDT, OT_RM_ALL, OT_NONE, 0x40002000 },
	/*0x0336*/{ 0x08, OI_STR, OT_RM16, OT_NONE, 0x40002000 },
	/*0x0337*/{ 0x08, OI_LLDT, OT_RM16, OT_NONE, 0x40002000 },
	/*0x0338*/{ 0x08, OI_LTR, OT_RM16, OT_NONE, 0x40012000 },
	/*0x0339*/{ 0x08, OI_VERR, OT_RM16, OT_NONE, 0x40002000 },
	/*0x033A*/{ 0x08, OI_VERW, OT_RM16, OT_NONE, 0x40002000 },
	/*0x033B*/{ 0x08, OI_SGDT, OT_M16_A_3264, OT_NONE, 0x44002000 },
	/*0x033C*/{ 0x08, OI_SIDT, OT_M16_A_3264, OT_NONE, 0x44002000 },
	/*0x033D*/{ 0x08, OI_LGDT, OT_M16_A_3264, OT_NONE, 0x44002000 },
	/*0x033E*/{ 0x08, OI_LIDT, OT_M16_A_3264, OT_NONE, 0x44002000 },
	/*0x033F*/{ 0x08, OI_SMSW, OT_R_ALL_M16, OT_NONE, 0xC0006000 },
	/*0x0340*/{ 0x08, OI_LMSW, OT_RM16, OT_NONE, 0x40006000 },
	/*0x0341*/{ 0x08, OI_INVLPG, OT_MEM, OT_NONE, 0x40012000 },
	/*0x0342*/{ 0x70, OI_VMCALL, OT_NONE, OT_NONE, 0x40010000 },
	/*0x0343*/{ 0x70, OI_VMLAUNCH, OT_NONE, OT_NONE, 0x40010000 },
	/*0x0344*/{ 0x70, OI_VMRESUME, OT_NONE, OT_NONE, 0x40010000 },
	/*0x0345*/{ 0x70, OI_VMXOFF, OT_NONE, OT_NONE, 0x40010000 },
	/*0x0346*/{ 0x08, OI_MONITOR, OT_NONE, OT_NONE, 0x40010000 },
	/*0x0347*/{ 0x08, OI_MWAIT, OT_NONE, OT_NONE, 0x40010000 },
	/*0x0348*/{ 0x08, OI_XGETBV, OT_NONE, OT_NONE, 0x00010000 },
	/*0x0349*/{ 0x08, OI_XSETBV, OT_NONE, OT_NONE, 0x00010000 },
	/*0x034A*/{ 0x70, OI_VMFUNC, OT_NONE, OT_NONE, 0x40010000 },
	/*0x034B*/{ 0x78, OI_VMRUN, OT_NONE, OT_NONE, 0x40012000 },
	/*0x034C*/{ 0x78, OI_VMMCALL, OT_NONE, OT_NONE, 0x40010000 },
	/*0x034D*/{ 0x78, OI_VMLOAD, OT_NONE, OT_NONE, 0x40010000 },
	/*0x034E*/{ 0x78, OI_VMSAVE, OT_NONE, OT_NONE, 0x40010000 },
	/*0x034F*/{ 0x78, OI_STGI, OT_NONE, OT_NONE, 0x40010000 },
	/*0x0350*/{ 0x78, OI_CLGI, OT_NONE, OT_NONE, 0x40010000 },
	/*0x0351*/{ 0x78, OI_SKINIT, OT_NONE, OT_NONE, 0x40010000 },
	/*0x0352*/{ 0x78, OI_INVLPGA, OT_NONE, OT_NONE, 0x40010000 },
	/*0x0353*/{ 0x08, OI_SWAPGS, OT_NONE, OT_NONE, 0x50000000 },
	/*0x0354*/{ 0x08, OI_RDTSCP, OT_NONE, OT_NONE, 0x50000000 },
	/*0x0355*/{ 0x60, OI_PREFETCH, OT_MEM, OT_NONE, 0x40012000 },
	/*0x0356*/{ 0x60, OI_PREFETCHW, OT_MEM, OT_NONE, 0x40012000 },
	/*0x0357*/{ 0x68, OI_PI2FW, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x0358*/{ 0x60, OI_PI2FD, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x0359*/{ 0x68, OI_PF2IW, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x035A*/{ 0x60, OI_PF2ID, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x035B*/{ 0x68, OI_PFNACC, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x035C*/{ 0x68, OI_PFPNACC, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x035D*/{ 0x60, OI_PFCMPGE, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x035E*/{ 0x60, OI_PFMIN, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x035F*/{ 0x60, OI_PFRCP, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x0360*/{ 0x60, OI_PFRSQRT, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x0361*/{ 0x60, OI_PFSUB, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x0362*/{ 0x60, OI_PFADD, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x0363*/{ 0x60, OI_PFCMPGT, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x0364*/{ 0x60, OI_PFMAX, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x0365*/{ 0x60, OI_PFRCPIT1, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x0366*/{ 0x60, OI_PFRSQIT1, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x0367*/{ 0x60, OI_PFSUBR, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x0368*/{ 0x60, OI_PFACC, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x0369*/{ 0x60, OI_PFCMPEQ, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x036A*/{ 0x60, OI_PFMUL, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x036B*/{ 0x60, OI_PFRCPIT2, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x036C*/{ 0x60, OI_PMULHRW, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x036D*/{ 0x68, OI_PSWAPD, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x036E*/{ 0x60, OI_PAVGUSB, OT_MMX, OT_MMX64, 0x00812000 },
	/*0x036F*/{ 0x28, OI_MOVUPS, OT_XMM, OT_XMM128, 0x80012000 },
	/*0x0370*/{ 0x30, OI_MOVUPD, OT_XMM, OT_XMM128, 0x80012000 },
	/*0x0371*/{ 0x28, OI_MOVSS, OT_XMM, OT_XMM32, 0x80012000 },
	/*0x0372*/{ 0x30, OI_MOVSD, OT_XMM, OT_XMM64, 0x80012000 },
	/*0x0373*/{ 0x28, OI_MOVUPS, OT_XMM128, OT_XMM, 0x80012000 },
	/*0x0374*/{ 0x30, OI_MOVUPD, OT_XMM128, OT_XMM, 0x80012000 },
	/*0x0375*/{ 0x28, OI_MOVSS, OT_XMM32, OT_XMM, 0x80012000 },
	/*0x0376*/{ 0x30, OI_MOVSD, OT_XMM64, OT_XMM, 0x80012000 },
	/*0x0377*/{ 0x30, OI_MOVLPD, OT_XMM, OT_MEM64, 0x80012000 },
	/*0x0378*/{ 0x38, OI_MOVSLDUP, OT_XMM, OT_XMM64, 0x80012000 },
	/*0x0379*/{ 0x38, OI_MOVDDUP, OT_XMM, OT_XMM64, 0x80012000 },
	/*0x037A*/{ 0x28, OI_MOVLPS, OT_MEM64, OT_XMM, 0x80012000 },
	/*0x037B*/{ 0x30, OI_MOVLPD, OT_MEM64, OT_XMM, 0x80012000 },
	/*0x037C*/{ 0x28, OI_UNPCKLPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x037D*/{ 0x30, OI_UNPCKLPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x037E*/{ 0x28, OI_UNPCKHPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x037F*/{ 0x30, OI_UNPCKHPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0380*/{ 0x30, OI_MOVHPD, OT_XMM, OT_MEM64, 0x80012000 },
	/*0x0381*/{ 0x38, OI_MOVSHDUP, OT_XMM, OT_XMM128, 0x80012000 },
	/*0x0382*/{ 0x28, OI_MOVHPS, OT_MEM64, OT_XMM, 0x80012000 },
	/*0x0383*/{ 0x30, OI_MOVHPD, OT_MEM64, OT_XMM, 0x80012000 },
	/*0x0384*/{ 0x28, OI_PREFETCHNTA, OT_MEM, OT_NONE, 0x40012000 },
	/*0x0385*/{ 0x28, OI_PREFETCHT0, OT_MEM, OT_NONE, 0x40012000 },
	/*0x0386*/{ 0x28, OI_PREFETCHT1, OT_MEM, OT_NONE, 0x40012000 },
	/*0x0387*/{ 0x28, OI_PREFETCHT2, OT_MEM, OT_NONE, 0x40012000 },
	/*0x0388*/{ 0x28, OI_MOVAPS, OT_XMM, OT_XMM128, 0x80012000 },
	/*0x0389*/{ 0x30, OI_MOVAPD, OT_XMM, OT_XMM128, 0x80012000 },
	/*0x038A*/{ 0x28, OI_MOVAPS, OT_XMM128, OT_XMM, 0x80012000 },
	/*0x038B*/{ 0x30, OI_MOVAPD, OT_XMM128, OT_XMM, 0x80012000 },
	/*0x038C*/{ 0x28, OI_CVTPI2PS, OT_XMM, OT_MMX64, 0x00012000 },
	/*0x038D*/{ 0x30, OI_CVTPI2PD, OT_XMM, OT_MMX64, 0x00012000 },
	/*0x038E*/{ 0x28, OI_CVTSI2SS, OT_XMM, OT_RM32_64, 0x04012800 },
	/*0x038F*/{ 0x30, OI_CVTSI2SD, OT_XMM, OT_RM32_64, 0x04012800 },
	/*0x0390*/{ 0x28, OI_MOVNTPS, OT_MEM128, OT_XMM, 0x80012000 },
	/*0x0391*/{ 0x30, OI_MOVNTPD, OT_MEM128, OT_XMM, 0x80012000 },
	/*0x0392*/{ 0x58, OI_MOVNTSS, OT_MEM32, OT_XMM, 0x80012000 },
	/*0x0393*/{ 0x58, OI_MOVNTSD, OT_MEM64, OT_XMM, 0x80012000 },
	/*0x0394*/{ 0x28, OI_CVTTPS2PI, OT_MMX, OT_XMM64, 0x00012000 },
	/*0x0395*/{ 0x30, OI_CVTTPD2PI, OT_MMX, OT_XMM128, 0x00012000 },
	/*0x0396*/{ 0x28, OI_CVTTSS2SI, OT_REG32_64, OT_XMM32, 0x04012800 },
	/*0x0397*/{ 0x30, OI_CVTTSD2SI, OT_REG32_64, OT_XMM64, 0x04012800 },
	/*0x0398*/{ 0x28, OI_CVTPS2PI, OT_MMX, OT_XMM64, 0x00012000 },
	/*0x0399*/{ 0x30, OI_CVTPD2PI, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x039A*/{ 0x28, OI_CVTSS2SI, OT_REG32_64, OT_XMM32, 0x04012800 },
	/*0x039B*/{ 0x30, OI_CVTSD2SI, OT_REG32_64, OT_XMM64, 0x04012800 },
	/*0x039C*/{ 0x28, OI_UCOMISS, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x039D*/{ 0x30, OI_UCOMISD, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x039E*/{ 0x28, OI_COMISS, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x039F*/{ 0x30, OI_COMISD, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x03A0*/{ 0x28, OI_MOVMSKPS, OT_REG32, OT_XMM_RM, 0x80412000 },
	/*0x03A1*/{ 0x30, OI_MOVMSKPD, OT_REG32, OT_XMM_RM, 0x80412000 },
	/*0x03A2*/{ 0x28, OI_SQRTPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03A3*/{ 0x30, OI_SQRTPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03A4*/{ 0x28, OI_SQRTSS, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x03A5*/{ 0x30, OI_SQRTSD, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x03A6*/{ 0x28, OI_RSQRTPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03A7*/{ 0x28, OI_RSQRTSS, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x03A8*/{ 0x28, OI_RCPPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03A9*/{ 0x28, OI_RCPSS, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x03AA*/{ 0x28, OI_ANDPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03AB*/{ 0x30, OI_ANDPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03AC*/{ 0x28, OI_ANDNPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03AD*/{ 0x30, OI_ANDNPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03AE*/{ 0x28, OI_ORPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03AF*/{ 0x30, OI_ORPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03B0*/{ 0x28, OI_XORPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03B1*/{ 0x30, OI_XORPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03B2*/{ 0x28, OI_ADDPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03B3*/{ 0x30, OI_ADDPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03B4*/{ 0x28, OI_ADDSS, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x03B5*/{ 0x30, OI_ADDSD, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x03B6*/{ 0x28, OI_MULPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03B7*/{ 0x30, OI_MULPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03B8*/{ 0x28, OI_MULSS, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x03B9*/{ 0x30, OI_MULSD, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x03BA*/{ 0x30, OI_CVTPS2PD, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x03BB*/{ 0x30, OI_CVTPD2PS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03BC*/{ 0x30, OI_CVTSS2SD, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x03BD*/{ 0x30, OI_CVTSD2SS, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x03BE*/{ 0x30, OI_CVTDQ2PS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03BF*/{ 0x30, OI_CVTPS2DQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03C0*/{ 0x30, OI_CVTTPS2DQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03C1*/{ 0x28, OI_SUBPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03C2*/{ 0x30, OI_SUBPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03C3*/{ 0x28, OI_SUBSS, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x03C4*/{ 0x30, OI_SUBSD, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x03C5*/{ 0x28, OI_MINPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03C6*/{ 0x30, OI_MINPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03C7*/{ 0x28, OI_MINSS, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x03C8*/{ 0x30, OI_MINSD, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x03C9*/{ 0x28, OI_DIVPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03CA*/{ 0x30, OI_DIVPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03CB*/{ 0x28, OI_DIVSS, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x03CC*/{ 0x30, OI_DIVSD, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x03CD*/{ 0x28, OI_MAXPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03CE*/{ 0x30, OI_MAXPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03CF*/{ 0x28, OI_MAXSS, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x03D0*/{ 0x30, OI_MAXSD, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x03D1*/{ 0x20, OI_PUNPCKLBW, OT_MMX, OT_MMX32, 0x00012000 },
	/*0x03D2*/{ 0x30, OI_PUNPCKLBW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03D3*/{ 0x20, OI_PUNPCKLWD, OT_MMX, OT_MMX32, 0x00012000 },
	/*0x03D4*/{ 0x30, OI_PUNPCKLWD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03D5*/{ 0x20, OI_PUNPCKLDQ, OT_MMX, OT_MMX32, 0x00012000 },
	/*0x03D6*/{ 0x30, OI_PUNPCKLDQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03D7*/{ 0x20, OI_PACKSSWB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x03D8*/{ 0x30, OI_PACKSSWB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03D9*/{ 0x20, OI_PCMPGTB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x03DA*/{ 0x30, OI_PCMPGTB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03DB*/{ 0x20, OI_PCMPGTW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x03DC*/{ 0x30, OI_PCMPGTW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03DD*/{ 0x20, OI_PCMPGTD, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x03DE*/{ 0x30, OI_PCMPGTD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03DF*/{ 0x20, OI_PACKUSWB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x03E0*/{ 0x30, OI_PACKUSWB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03E1*/{ 0x20, OI_PUNPCKHBW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x03E2*/{ 0x30, OI_PUNPCKHBW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03E3*/{ 0x20, OI_PUNPCKHWD, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x03E4*/{ 0x30, OI_PUNPCKHWD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03E5*/{ 0x20, OI_PUNPCKHDQ, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x03E6*/{ 0x30, OI_PUNPCKHDQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03E7*/{ 0x20, OI_PACKSSDW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x03E8*/{ 0x30, OI_PACKSSDW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03E9*/{ 0x30, OI_PUNPCKLQDQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03EA*/{ 0x30, OI_PUNPCKHQDQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03EB*/{ 0x20, OI_MOVQ, OT_MMX, OT_MMX64, 0x80012000 },
	/*0x03EC*/{ 0x30, OI_MOVDQA, OT_XMM, OT_XMM128, 0x80012000 },
	/*0x03ED*/{ 0x30, OI_MOVDQU, OT_XMM, OT_XMM128, 0x80012000 },
	/*0x03EE*/{ 0x20, OI_PCMPEQB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x03EF*/{ 0x30, OI_PCMPEQB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03F0*/{ 0x20, OI_PCMPEQW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x03F1*/{ 0x30, OI_PCMPEQW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03F2*/{ 0x20, OI_PCMPEQD, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x03F3*/{ 0x30, OI_PCMPEQD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03F4*/{ 0x20, OI_EMMS, OT_NONE, OT_NONE, 0x00010000 },
	/*0x03F5*/{ 0x70, OI_VMREAD, OT_RM32_64, OT_REG32_64, 0x04012000 },
	/*0x03F6*/{ 0x70, OI_VMWRITE, OT_REG32_64, OT_RM32_64, 0x04012000 },
	/*0x03F7*/{ 0x58, OI_EXTRQ, OT_XMM, OT_XMM_RM, 0x00012000 },
	/*0x03F8*/{ 0x58, OI_INSERTQ, OT_XMM, OT_XMM_RM, 0x00412000 },
	/*0x03F9*/{ 0x70, OI_CVTPH2PS, OT_XMM, OT_XMM64, 0x00010000 },
	/*0x03FA*/{ 0x70, OI_CVTPS2PH, OT_XMM64, OT_XMM, 0x00010000 },
	/*0x03FB*/{ 0x38, OI_HADDPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03FC*/{ 0x38, OI_HADDPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03FD*/{ 0x38, OI_HSUBPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03FE*/{ 0x38, OI_HSUBPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x03FF*/{ 0x30, OI_MOVQ, OT_XMM, OT_XMM64, 0x80012000 },
	/*0x0400*/{ 0x20, OI_MOVQ, OT_MMX64, OT_MMX, 0x80012000 },
	/*0x0401*/{ 0x30, OI_MOVDQA, OT_XMM128, OT_XMM, 0x80012000 },
	/*0x0402*/{ 0x30, OI_MOVDQU, OT_XMM128, OT_XMM, 0x80012000 },
	/*0x0403*/{ 0x50, OI_POPCNT, OT_REG_ALL, OT_RM_ALL, 0x80012200 },
	/*0x0404*/{ 0x08, OI_BT, OT_RM_ALL, OT_IMM8, 0xC0012000 },
	/*0x0405*/{ 0x08, OI_BTS, OT_RM_ALL, OT_IMM8, 0xC0012001 },
	/*0x0406*/{ 0x08, OI_BTR, OT_RM_ALL, OT_IMM8, 0xC0012001 },
	/*0x0407*/{ 0x08, OI_BTC, OT_RM_ALL, OT_IMM8, 0xC0012001 },
	/*0x0408*/{ 0x08, OI_BSF, OT_REG_ALL, OT_RM_ALL, 0x00012000 },
	/*0x0409*/{ 0xA1, OI_TZCNT, OT_REG_ALL, OT_RM_ALL, 0x00002200 },
	/*0x040A*/{ 0x08, OI_BSR, OT_REG_ALL, OT_RM_ALL, 0x00012000 },
	/*0x040B*/{ 0xA1, OI_LZCNT, OT_REG_ALL, OT_RM_ALL, 0x80002200 },
	/*0x040C*/{ 0x70, OI_VMPTRST, OT_MEM64, OT_NONE, 0x40012000 },
	/*0x040D*/{ 0x38, OI_ADDSUBPD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x040E*/{ 0x38, OI_ADDSUBPS, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x040F*/{ 0x20, OI_PSRLW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0410*/{ 0x30, OI_PSRLW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0411*/{ 0x20, OI_PSRLD, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0412*/{ 0x30, OI_PSRLD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0413*/{ 0x20, OI_PSRLQ, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0414*/{ 0x30, OI_PSRLQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0415*/{ 0x30, OI_PADDQ, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x0416*/{ 0x30, OI_PADDQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0417*/{ 0x20, OI_PMULLW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0418*/{ 0x30, OI_PMULLW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0419*/{ 0x30, OI_MOVQ, OT_XMM64, OT_XMM, 0x80012000 },
	/*0x041A*/{ 0x30, OI_MOVQ2DQ, OT_XMM, OT_MMX_RM, 0x80412000 },
	/*0x041B*/{ 0x30, OI_MOVDQ2Q, OT_MMX, OT_XMM_RM, 0x80412000 },
	/*0x041C*/{ 0x28, OI_PMOVMSKB, OT_REG32, OT_MMX_RM, 0x00412000 },
	/*0x041D*/{ 0x30, OI_PMOVMSKB, OT_REG32, OT_XMM_RM, 0x00412000 },
	/*0x041E*/{ 0x20, OI_PSUBUSB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x041F*/{ 0x30, OI_PSUBUSB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0420*/{ 0x20, OI_PSUBUSW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0421*/{ 0x30, OI_PSUBUSW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0422*/{ 0x28, OI_PMINUB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0423*/{ 0x30, OI_PMINUB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0424*/{ 0x20, OI_PAND, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0425*/{ 0x30, OI_PAND, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0426*/{ 0x20, OI_PADDUSB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0427*/{ 0x30, OI_PADDUSB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0428*/{ 0x20, OI_PADDUSW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0429*/{ 0x30, OI_PADDUSW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x042A*/{ 0x28, OI_PMAXUB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x042B*/{ 0x30, OI_PMAXUB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x042C*/{ 0x20, OI_PANDN, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x042D*/{ 0x30, OI_PANDN, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x042E*/{ 0x28, OI_PAVGB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x042F*/{ 0x30, OI_PAVGB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0430*/{ 0x20, OI_PSRAW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0431*/{ 0x30, OI_PSRAW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0432*/{ 0x20, OI_PSRAD, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0433*/{ 0x30, OI_PSRAD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0434*/{ 0x28, OI_PAVGW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0435*/{ 0x30, OI_PAVGW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0436*/{ 0x28, OI_PMULHUW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0437*/{ 0x30, OI_PMULHUW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0438*/{ 0x20, OI_PMULHW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0439*/{ 0x30, OI_PMULHW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x043A*/{ 0x30, OI_CVTTPD2DQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x043B*/{ 0x30, OI_CVTDQ2PD, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x043C*/{ 0x30, OI_CVTPD2DQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x043D*/{ 0x28, OI_MOVNTQ, OT_MEM64, OT_MMX, 0x80012000 },
	/*0x043E*/{ 0x30, OI_MOVNTDQ, OT_MEM128, OT_XMM, 0x80012000 },
	/*0x043F*/{ 0x20, OI_PSUBSB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0440*/{ 0x30, OI_PSUBSB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0441*/{ 0x20, OI_PSUBSW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0442*/{ 0x30, OI_PSUBSW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0443*/{ 0x28, OI_PMINSW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0444*/{ 0x30, OI_PMINSW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0445*/{ 0x20, OI_POR, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0446*/{ 0x30, OI_POR, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0447*/{ 0x20, OI_PADDSB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0448*/{ 0x30, OI_PADDSB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0449*/{ 0x20, OI_PADDSW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x044A*/{ 0x30, OI_PADDSW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x044B*/{ 0x28, OI_PMAXSW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x044C*/{ 0x30, OI_PMAXSW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x044D*/{ 0x20, OI_PXOR, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x044E*/{ 0x30, OI_PXOR, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x044F*/{ 0x38, OI_LDDQU, OT_XMM, OT_MEM, 0x00012000 },
	/*0x0450*/{ 0x20, OI_PSLLW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0451*/{ 0x30, OI_PSLLW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0452*/{ 0x20, OI_PSLLD, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0453*/{ 0x30, OI_PSLLD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0454*/{ 0x20, OI_PSLLQ, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0455*/{ 0x30, OI_PSLLQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0456*/{ 0x30, OI_PMULUDQ, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0457*/{ 0x30, OI_PMULUDQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0458*/{ 0x20, OI_PMADDWD, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0459*/{ 0x30, OI_PMADDWD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x045A*/{ 0x28, OI_PSADBW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x045B*/{ 0x30, OI_PSADBW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x045C*/{ 0x28, OI_MASKMOVQ, OT_MMX, OT_MMX_RM, 0x00412000 },
	/*0x045D*/{ 0x30, OI_MASKMOVDQU, OT_XMM, OT_XMM_RM, 0x00412000 },
	/*0x045E*/{ 0x20, OI_PSUBB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x045F*/{ 0x30, OI_PSUBB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0460*/{ 0x20, OI_PSUBW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0461*/{ 0x30, OI_PSUBW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0462*/{ 0x20, OI_PSUBD, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0463*/{ 0x30, OI_PSUBD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0464*/{ 0x30, OI_PSUBQ, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0465*/{ 0x30, OI_PSUBQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0466*/{ 0x20, OI_PADDB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0467*/{ 0x30, OI_PADDB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0468*/{ 0x20, OI_PADDW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0469*/{ 0x30, OI_PADDW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x046A*/{ 0x20, OI_PADDD, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x046B*/{ 0x30, OI_PADDD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x046C*/{ 0x10, OI_FNSTENV, OT_MEM, OT_NONE, 0xC0002000 },
	/*0x046D*/{ 0x10, OI_FSTENV, OT_MEM, OT_NONE, 0xC0012200 },
	/*0x046E*/{ 0x10, OI_FNSTCW, OT_FPUM16, OT_NONE, 0xC0002000 },
	/*0x046F*/{ 0x10, OI_FSTCW, OT_FPUM16, OT_NONE, 0xC0012200 },
	/*0x0470*/{ 0x10, OI_FNCLEX, OT_NONE, OT_NONE, 0x40000000 },
	/*0x0471*/{ 0x10, OI_FCLEX, OT_NONE, OT_NONE, 0x40010200 },
	/*0x0472*/{ 0x10, OI_FNINIT, OT_NONE, OT_NONE, 0x40000000 },
	/*0x0473*/{ 0x10, OI_FINIT, OT_NONE, OT_NONE, 0x40010200 },
	/*0x0474*/{ 0x10, OI_FNSAVE, OT_MEM, OT_NONE, 0xC0002000 },
	/*0x0475*/{ 0x10, OI_FSAVE, OT_MEM, OT_NONE, 0xC0012200 },
	/*0x0476*/{ 0x10, OI_FNSTSW, OT_FPUM16, OT_NONE, 0xC0002000 },
	/*0x0477*/{ 0x10, OI_FSTSW, OT_FPUM16, OT_NONE, 0xC0012200 },
	/*0x0478*/{ 0x10, OI_FNSTSW, OT_ACC16, OT_NONE, 0xC0000000 },
	/*0x0479*/{ 0x10, OI_FSTSW, OT_ACC16, OT_NONE, 0xC0010200 },
	/*0x047A*/{ 0x40, OI_PSHUFB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x047B*/{ 0x40, OI_PSHUFB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x047C*/{ 0x40, OI_PHADDW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x047D*/{ 0x40, OI_PHADDW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x047E*/{ 0x40, OI_PHADDD, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x047F*/{ 0x40, OI_PHADDD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0480*/{ 0x40, OI_PHADDSW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0481*/{ 0x40, OI_PHADDSW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0482*/{ 0x40, OI_PMADDUBSW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0483*/{ 0x40, OI_PMADDUBSW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0484*/{ 0x40, OI_PHSUBW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0485*/{ 0x40, OI_PHSUBW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0486*/{ 0x40, OI_PHSUBD, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0487*/{ 0x40, OI_PHSUBD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0488*/{ 0x40, OI_PHSUBSW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0489*/{ 0x40, OI_PHSUBSW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x048A*/{ 0x40, OI_PSIGNB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x048B*/{ 0x40, OI_PSIGNB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x048C*/{ 0x40, OI_PSIGNW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x048D*/{ 0x40, OI_PSIGNW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x048E*/{ 0x40, OI_PSIGND, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x048F*/{ 0x40, OI_PSIGND, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0490*/{ 0x40, OI_PMULHRSW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0491*/{ 0x40, OI_PMULHRSW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0492*/{ 0x50, OI_PTEST, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0493*/{ 0x40, OI_PABSB, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0494*/{ 0x40, OI_PABSB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0495*/{ 0x40, OI_PABSW, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0496*/{ 0x40, OI_PABSW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0497*/{ 0x40, OI_PABSD, OT_MMX, OT_MMX64, 0x00012000 },
	/*0x0498*/{ 0x40, OI_PABSD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x0499*/{ 0x48, OI_PMOVSXBW, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x049A*/{ 0x48, OI_PMOVSXBD, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x049B*/{ 0x48, OI_PMOVSXBQ, OT_XMM, OT_XMM16, 0x00012000 },
	/*0x049C*/{ 0x48, OI_PMOVSXWD, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x049D*/{ 0x48, OI_PMOVSXWQ, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x049E*/{ 0x48, OI_PMOVSXDQ, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x049F*/{ 0x48, OI_PMULDQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04A0*/{ 0x48, OI_PCMPEQQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04A1*/{ 0x48, OI_MOVNTDQA, OT_XMM, OT_MEM128, 0x80012000 },
	/*0x04A2*/{ 0x48, OI_PACKUSDW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04A3*/{ 0x48, OI_PMOVZXBW, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x04A4*/{ 0x48, OI_PMOVZXBD, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x04A5*/{ 0x48, OI_PMOVZXBQ, OT_XMM, OT_XMM16, 0x00012000 },
	/*0x04A6*/{ 0x48, OI_PMOVZXWD, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x04A7*/{ 0x48, OI_PMOVZXWQ, OT_XMM, OT_XMM32, 0x00012000 },
	/*0x04A8*/{ 0x48, OI_PMOVZXDQ, OT_XMM, OT_XMM64, 0x00012000 },
	/*0x04A9*/{ 0x50, OI_PCMPGTQ, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04AA*/{ 0x48, OI_PMINSB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04AB*/{ 0x48, OI_PMINSD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04AC*/{ 0x48, OI_PMINUW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04AD*/{ 0x48, OI_PMINUD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04AE*/{ 0x48, OI_PMAXSB, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04AF*/{ 0x48, OI_PMAXSD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04B0*/{ 0x48, OI_PMAXUW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04B1*/{ 0x48, OI_PMAXUD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04B2*/{ 0x48, OI_PMULLD, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04B3*/{ 0x48, OI_PHMINPOSUW, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04B4*/{ 0x70, OI_INVEPT, OT_REG32_64, OT_MEM128, 0x04012000 },
	/*0x04B5*/{ 0x70, OI_INVVPID, OT_REG32_64, OT_MEM128, 0x04012000 },
	/*0x04B6*/{ 0x70, OI_INVPCID, OT_REG32_64, OT_MEM128, 0x04012000 },
	/*0x04B7*/{ 0x98, OI_AESIMC, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04B8*/{ 0x98, OI_AESENC, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04B9*/{ 0x98, OI_AESENCLAST, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04BA*/{ 0x98, OI_AESDEC, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04BB*/{ 0x98, OI_AESDECLAST, OT_XMM, OT_XMM128, 0x00012000 },
	/*0x04BC*/{ 0x08, OI_MOVBE, OT_REG_ALL, OT_RM_ALL, 0x80012000 },
	/*0x04BD*/{ 0x50, OI_CRC32, OT_REG32_64, OT_RM8, 0x80012000 },
	/*0x04BE*/{ 0x08, OI_MOVBE, OT_RM_ALL, OT_REG_ALL, 0x80012000 },
	/*0x04BF*/{ 0x50, OI_CRC32, OT_REG32_64, OT_RM_ALL, 0x80012000 },
	/*0x04C0*/{ 0x20, OI_PSRLW, OT_MMX_RM, OT_IMM8, 0x40412000 },
	/*0x04C1*/{ 0x30, OI_PSRLW, OT_XMM_RM, OT_IMM8, 0x40012000 },
	/*0x04C2*/{ 0x20, OI_PSRAW, OT_MMX_RM, OT_IMM8, 0x40412000 },
	/*0x04C3*/{ 0x30, OI_PSRAW, OT_XMM_RM, OT_IMM8, 0x40012000 },
	/*0x04C4*/{ 0x20, OI_PSLLW, OT_MMX_RM, OT_IMM8, 0x40412000 },
	/*0x04C5*/{ 0x30, OI_PSLLW, OT_XMM_RM, OT_IMM8, 0x40012000 },
	/*0x04C6*/{ 0x20, OI_PSRLD, OT_MMX_RM, OT_IMM8, 0x40412000 },
	/*0x04C7*/{ 0x30, OI_PSRLD, OT_XMM_RM, OT_IMM8, 0x40012000 },
	/*0x04C8*/{ 0x20, OI_PSRAD, OT_MMX_RM, OT_IMM8, 0x40412000 },
	/*0x04C9*/{ 0x30, OI_PSRAD, OT_XMM_RM, OT_IMM8, 0x40012000 },
	/*0x04CA*/{ 0x20, OI_PSLLD, OT_MMX_RM, OT_IMM8, 0x40412000 },
	/*0x04CB*/{ 0x30, OI_PSLLD, OT_XMM_RM, OT_IMM8, 0x40012000 },
	/*0x04CC*/{ 0x20, OI_PSRLQ, OT_MMX_RM, OT_IMM8, 0x40412000 },
	/*0x04CD*/{ 0x30, OI_PSRLQ, OT_XMM_RM, OT_IMM8, 0x40012000 },
	/*0x04CE*/{ 0x30, OI_PSRLDQ, OT_XMM_RM, OT_IMM8, 0x40012000 },
	/*0x04CF*/{ 0x20, OI_PSLLQ, OT_MMX_RM, OT_IMM8, 0x40412000 },
	/*0x04D0*/{ 0x30, OI_PSLLQ, OT_XMM_RM, OT_IMM8, 0x40012000 },
	/*0x04D1*/{ 0x30, OI_PSLLDQ, OT_XMM_RM, OT_IMM8, 0x40012000 },
	/*0x04D2*/{ 0x70, OI_RDFSBASE, OT_REG32_64, OT_NONE, 0x44012000 },
	/*0x04D3*/{ 0x70, OI_RDGSBASE, OT_REG32_64, OT_NONE, 0x44012000 },
	/*0x04D4*/{ 0x08, OI_LDMXCSR, OT_MEM32, OT_NONE, 0x40012000 },
	/*0x04D5*/{ 0x70, OI_WRFSBASE, OT_REG32_64, OT_NONE, 0x44012000 },
	/*0x04D6*/{ 0x08, OI_STMXCSR, OT_MEM32, OT_NONE, 0x40012000 },
	/*0x04D7*/{ 0x70, OI_WRGSBASE, OT_REG32_64, OT_NONE, 0x44012000 },
	/*0x04D8*/{ 0x70, OI_VMPTRLD, OT_MEM, OT_NONE, 0x40012000 },
	/*0x04D9*/{ 0x70, OI_VMCLEAR, OT_MEM64, OT_NONE, 0x40012000 },
	/*0x04DA*/{ 0x70, OI_VMXON, OT_MEM64, OT_NONE, 0x40012000 },
	/*0x04DB*/{ 0x08, OI_WAIT, OT_NONE, OT_NONE, 0x00000000 },
	/*0x04DC*/{ 0xA1, OI_ADCX, OT_REG32_64, OT_RM32_64, 0x80013000 },
	/*0x04DD*/{ 0xA1, OI_ADOX, OT_REG32_64, OT_RM32_64, 0x80013000 },
	/*0x04DE*/{ 0x08, OI_LFENCE, OT_NONE, OT_NONE, 0x40012000 },
	/*0x04DF*/{ 0x08, OI_MFENCE, OT_NONE, OT_NONE, 0x40012000 },
	/*0x04E0*/{ 0x08, OI_SFENCE, OT_NONE, OT_NONE, 0x40012000 },
	/*0x04E1*/{ 0x08, OI_RDRAND, OT_RM_ALL, OT_NONE, 0x40002801 },
	/*0x04E2*/{ 0x08, OI_RDSEED, OT_RM_ALL, OT_NONE, 0x40002801 },
};

OP_INFOEX PEDisAsm::OP_INFOEXS[OP_INFOEX_SIZE] =
{ //418
	/*0x0000*/{ 0x08, OI_IMUL, OT_REG_ALL, OT_RM_ALL, 0x80082000, OT_IMM_ALL, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0001*/{ 0x08, OI_IMUL, OT_REG_ALL, OT_RM_ALL, 0x80082000, OT_SE_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0002*/{ 0x08, OI_CBW, OT_NONE, OT_NONE, 0x08040000, OT_NONE, OT_NONE, 0x00, OI_CWDE, OI_CDQE },
	/*0x0003*/{ 0x08, OI_CWD, OT_NONE, OT_NONE, 0x08040000, OT_NONE, OT_NONE, 0x00, OI_CDQ, OI_CQO },
	/*0x0004*/{ 0x0D, OI_JCXZ, OT_REL_8, OT_NONE, 0x08040400, OT_NONE, OT_NONE, 0x00, OI_JECXZ, OI_JRCXZ },
	/*0x0005*/{ 0x08, OI_SHLD, OT_RM_ALL, OT_REG_ALL, 0x80092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0006*/{ 0x08, OI_SHLD, OT_RM_ALL, OT_REG_ALL, 0x80092000, OT_REG_CL, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0007*/{ 0x08, OI_SHRD, OT_RM_ALL, OT_REG_ALL, 0x80092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0008*/{ 0x08, OI_SHRD, OT_RM_ALL, OT_REG_ALL, 0x80092000, OT_REG_CL, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0009*/{ 0x80, OI_VMOVUPS, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x000A*/{ 0x80, OI_VMOVUPD, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x000B*/{ 0x80, OI_VMOVSS, OT_XMM, OT_VXMM, 0x40093000, OT_XMM_RM, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x000C*/{ 0x80, OI_VMOVSD, OT_XMM, OT_VXMM, 0x40093000, OT_XMM_RM, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x000D*/{ 0x80, OI_VMOVSS, OT_XMM, OT_MEM32, 0x40013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x000E*/{ 0x80, OI_VMOVSD, OT_XMM, OT_MEM64, 0x40013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x000F*/{ 0x80, OI_VMOVUPS, OT_YXMM128_256, OT_YXMM, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0010*/{ 0x80, OI_VMOVUPD, OT_YXMM128_256, OT_YXMM, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0011*/{ 0x80, OI_VMOVSS, OT_XMM, OT_VXMM, 0x40093000, OT_XMM_RM, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0012*/{ 0x80, OI_VMOVSD, OT_XMM, OT_VXMM, 0x40093000, OT_XMM_RM, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0013*/{ 0x80, OI_VMOVSS, OT_MEM32, OT_XMM, 0x40013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0014*/{ 0x80, OI_VMOVSD, OT_MEM64, OT_XMM, 0x40013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0015*/{ 0x28, OI_MOVHLPS, OT_XMM, OT_XMM64, 0x80252000, OT_NONE, OT_NONE, 0x00, OI_MOVLPS, OI_UNKNOWN },
	/*0x0016*/{ 0x80, OI_VMOVHLPS, OT_XMM, OT_VXMM, 0x002D3000, OT_XMM64, OT_NONE, 0x00, OI_VMOVLPS, OI_UNKNOWN },
	/*0x0017*/{ 0x80, OI_VMOVLPD, OT_XMM, OT_VXMM, 0x00093000, OT_MEM64, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0018*/{ 0x80, OI_VMOVSLDUP, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0019*/{ 0x80, OI_VMOVDDUP, OT_YXMM, OT_YXMM64_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x001A*/{ 0x80, OI_VMOVLPS, OT_MEM64, OT_XMM, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x001B*/{ 0x80, OI_VMOVLPD, OT_MEM64, OT_XMM, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x001C*/{ 0x80, OI_VUNPCKLPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x001D*/{ 0x80, OI_VUNPCKLPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x001E*/{ 0x80, OI_VUNPCKHPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x001F*/{ 0x80, OI_VUNPCKHPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0020*/{ 0x28, OI_MOVLHPS, OT_XMM, OT_XMM64, 0x80252000, OT_NONE, OT_NONE, 0x00, OI_MOVHPS, OI_UNKNOWN },
	/*0x0021*/{ 0x80, OI_VMOVLHPS, OT_XMM, OT_VXMM, 0x002D3000, OT_XMM64, OT_NONE, 0x00, OI_VMOVHPS, OI_UNKNOWN },
	/*0x0022*/{ 0x80, OI_VMOVHPD, OT_XMM, OT_VXMM, 0x00093000, OT_MEM64, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0023*/{ 0x80, OI_VMOVSHDUP, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0024*/{ 0x80, OI_VMOVHPS, OT_MEM64, OT_XMM, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0025*/{ 0x80, OI_VMOVHPD, OT_MEM64, OT_XMM, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0026*/{ 0x80, OI_VMOVAPS, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0027*/{ 0x80, OI_VMOVAPD, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0028*/{ 0x80, OI_VMOVAPS, OT_YXMM128_256, OT_YXMM, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0029*/{ 0x80, OI_VMOVAPD, OT_YXMM128_256, OT_YXMM, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x002A*/{ 0x80, OI_VCVTSI2SS, OT_XMM, OT_VXMM, 0x00093000, OT_WRM32_64, OT_NONE, 0x02, OI_UNKNOWN, OI_UNKNOWN },
	/*0x002B*/{ 0x80, OI_VCVTSI2SD, OT_XMM, OT_VXMM, 0x00093000, OT_WRM32_64, OT_NONE, 0x02, OI_UNKNOWN, OI_UNKNOWN },
	/*0x002C*/{ 0x80, OI_VMOVNTPS, OT_LMEM128_256, OT_YXMM, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x002D*/{ 0x80, OI_VMOVNTPD, OT_LMEM128_256, OT_YXMM, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x002E*/{ 0x80, OI_VCVTTSS2SI, OT_WREG32_64, OT_WXMM32_64, 0x00013000, OT_NONE, OT_NONE, 0x42, OI_UNKNOWN, OI_UNKNOWN },
	/*0x002F*/{ 0x80, OI_VCVTTSD2SI, OT_WREG32_64, OT_WXMM32_64, 0x00013000, OT_NONE, OT_NONE, 0x42, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0030*/{ 0x80, OI_VCVTSS2SI, OT_WREG32_64, OT_WXMM32_64, 0x00013000, OT_NONE, OT_NONE, 0x42, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0031*/{ 0x80, OI_VCVTSD2SI, OT_WREG32_64, OT_WXMM32_64, 0x00013000, OT_NONE, OT_NONE, 0x42, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0032*/{ 0x80, OI_VUCOMISS, OT_XMM, OT_XMM32, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0033*/{ 0x80, OI_VUCOMISD, OT_XMM, OT_XMM64, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0034*/{ 0x80, OI_VCOMISS, OT_XMM, OT_XMM32, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0035*/{ 0x80, OI_VCOMISD, OT_XMM, OT_XMM64, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0036*/{ 0x80, OI_VMOVMSKPS, OT_REG32_64, OT_YXMM, 0x04413000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0037*/{ 0x80, OI_VMOVMSKPD, OT_REG32_64, OT_YXMM, 0x04413000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0038*/{ 0x80, OI_VSQRTPS, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0039*/{ 0x80, OI_VSQRTPD, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x003A*/{ 0x80, OI_VSQRTSS, OT_XMM, OT_VXMM, 0x00093000, OT_XMM32, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x003B*/{ 0x80, OI_VSQRTSD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM64, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x003C*/{ 0x80, OI_VRSQRTPS, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x003D*/{ 0x80, OI_VRSQRTSS, OT_XMM, OT_VXMM, 0x00093000, OT_XMM32, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x003E*/{ 0x80, OI_VRCPPS, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x003F*/{ 0x80, OI_VRCPSS, OT_XMM, OT_VXMM, 0x00093000, OT_XMM32, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0040*/{ 0x80, OI_VANDPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0041*/{ 0x80, OI_VANDPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0042*/{ 0x80, OI_VANDNPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0043*/{ 0x80, OI_VANDNPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0044*/{ 0x80, OI_VORPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0045*/{ 0x80, OI_VORPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0046*/{ 0x80, OI_VXORPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0047*/{ 0x80, OI_VXORPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0048*/{ 0x80, OI_VADDPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0049*/{ 0x80, OI_VADDPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x004A*/{ 0x80, OI_VADDSS, OT_XMM, OT_VXMM, 0x00093000, OT_XMM32, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x004B*/{ 0x80, OI_VADDSD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM64, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x004C*/{ 0x80, OI_VMULPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x004D*/{ 0x80, OI_VMULPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x004E*/{ 0x80, OI_VMULSS, OT_XMM, OT_VXMM, 0x00093000, OT_XMM32, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x004F*/{ 0x80, OI_VMULSD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM64, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0050*/{ 0x80, OI_VCVTPS2PD, OT_YXMM, OT_LXMM64_128, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0051*/{ 0x80, OI_VCVTPD2PS, OT_XMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0052*/{ 0x80, OI_VCVTSS2SD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM32, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0053*/{ 0x80, OI_VCVTSD2SS, OT_XMM, OT_VXMM, 0x00093000, OT_XMM64, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0054*/{ 0x80, OI_VCVTDQ2PS, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0055*/{ 0x80, OI_VCVTPS2DQ, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0056*/{ 0x80, OI_VCVTTPS2DQ, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0057*/{ 0x80, OI_VSUBPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0058*/{ 0x80, OI_VSUBPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0059*/{ 0x80, OI_VSUBSS, OT_XMM, OT_VXMM, 0x00093000, OT_XMM32, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x005A*/{ 0x80, OI_VSUBSD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM64, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x005B*/{ 0x80, OI_VMINPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x005C*/{ 0x80, OI_VMINPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x005D*/{ 0x80, OI_VMINSS, OT_XMM, OT_VXMM, 0x00093000, OT_XMM32, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x005E*/{ 0x80, OI_VMINSD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM64, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x005F*/{ 0x80, OI_VDIVPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0060*/{ 0x80, OI_VDIVPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0061*/{ 0x80, OI_VDIVSS, OT_XMM, OT_VXMM, 0x00093000, OT_XMM32, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0062*/{ 0x80, OI_VDIVSD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM64, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0063*/{ 0x80, OI_VMAXPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0064*/{ 0x80, OI_VMAXPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0065*/{ 0x80, OI_VMAXSS, OT_XMM, OT_VXMM, 0x00093000, OT_XMM32, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0066*/{ 0x80, OI_VMAXSD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM64, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0067*/{ 0x80, OI_VPUNPCKLBW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0068*/{ 0x80, OI_VPUNPCKLWD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0069*/{ 0x80, OI_VPUNPCKLDQ, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x006A*/{ 0x80, OI_VPACKSSWB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x006B*/{ 0x80, OI_VPCMPGTB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x006C*/{ 0x80, OI_VPCMPGTW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x006D*/{ 0x80, OI_VPCMPGTD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x006E*/{ 0x80, OI_VPACKUSWB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x006F*/{ 0x80, OI_VPUNPCKHBW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0070*/{ 0x80, OI_VPUNPCKHWD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0071*/{ 0x80, OI_VPUNPCKHDQ, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0072*/{ 0x80, OI_VPACKSSDW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0073*/{ 0x80, OI_VPUNPCKLQDQ, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0074*/{ 0x80, OI_VPUNPCKHQDQ, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0075*/{ 0x20, OI_MOVD, OT_MMX, OT_RM32_64, 0x8C012800, OT_NONE, OT_NONE, 0x00, OI_UNKNOWN, OI_MOVQ },
	/*0x0076*/{ 0x30, OI_MOVD, OT_XMM, OT_RM32_64, 0x8C012800, OT_NONE, OT_NONE, 0x00, OI_UNKNOWN, OI_MOVQ },
	/*0x0077*/{ 0x80, OI_VMOVD, OT_XMM, OT_WRM32_64, 0x00013000, OT_NONE, OT_NONE, 0x46, OI_VMOVQ, OI_UNKNOWN },
	/*0x0078*/{ 0x80, OI_VMOVDQA, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0079*/{ 0x80, OI_VMOVDQU, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x007A*/{ 0x28, OI_PSHUFW, OT_MMX, OT_MMX64, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x007B*/{ 0x30, OI_PSHUFD, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x007C*/{ 0x30, OI_PSHUFHW, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x007D*/{ 0x30, OI_PSHUFLW, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x007E*/{ 0x80, OI_VPSHUFD, OT_XMM, OT_XMM128, 0x00093000, OT_IMM8, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x007F*/{ 0x80, OI_VPSHUFHW, OT_XMM, OT_XMM128, 0x00093000, OT_IMM8, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0080*/{ 0x80, OI_VPSHUFLW, OT_XMM, OT_XMM128, 0x00093000, OT_IMM8, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0081*/{ 0x80, OI_VPCMPEQB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0082*/{ 0x80, OI_VPCMPEQW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0083*/{ 0x80, OI_VPCMPEQD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0084*/{ 0x80, OI_VZEROUPPER, OT_NONE, OT_NONE, 0x00011000, OT_NONE, OT_NONE, 0x49, OI_VZEROALL, OI_UNKNOWN },
	/*0x0085*/{ 0x58, OI_EXTRQ, OT_XMM_RM, OT_IMM8, 0x20092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0086*/{ 0x58, OI_INSERTQ, OT_XMM, OT_XMM_RM, 0x00592000, OT_IMM8, OT_IMM8, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0087*/{ 0x80, OI_VHADDPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0088*/{ 0x80, OI_VHADDPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0089*/{ 0x80, OI_VHSUBPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x008A*/{ 0x80, OI_VHSUBPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x008B*/{ 0x20, OI_MOVD, OT_RM32_64, OT_MMX, 0x8C012800, OT_NONE, OT_NONE, 0x00, OI_UNKNOWN, OI_MOVQ },
	/*0x008C*/{ 0x30, OI_MOVD, OT_RM32_64, OT_XMM, 0x8C012800, OT_NONE, OT_NONE, 0x00, OI_UNKNOWN, OI_MOVQ },
	/*0x008D*/{ 0x80, OI_VMOVD, OT_WRM32_64, OT_XMM, 0x00013000, OT_NONE, OT_NONE, 0x46, OI_VMOVQ, OI_UNKNOWN },
	/*0x008E*/{ 0x80, OI_VMOVQ, OT_XMM, OT_XMM64, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x008F*/{ 0x80, OI_VMOVDQA, OT_YXMM128_256, OT_YXMM, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0090*/{ 0x80, OI_VMOVDQU, OT_YXMM128_256, OT_YXMM, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0091*/{ 0x08, OI_XSAVE, OT_MEM, OT_NONE, 0x4C012800, OT_NONE, OT_NONE, 0x00, OI_UNKNOWN, OI_XSAVE64 },
	/*0x0092*/{ 0x08, OI_XRSTOR, OT_MEM, OT_NONE, 0x4C012800, OT_NONE, OT_NONE, 0x00, OI_UNKNOWN, OI_XRSTOR64 },
	/*0x0093*/{ 0x08, OI_XSAVEOPT, OT_MEM, OT_NONE, 0x4C012800, OT_NONE, OT_NONE, 0x00, OI_UNKNOWN, OI_XSAVEOPT64 },
	/*0x0094*/{ 0x08, OI_CLFLUSH, OT_MEM, OT_NONE, 0x40012000, OT_NONE, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0095*/{ 0x28, OI_CMPEQPS, OT_XMM, OT_XMM128, 0x01052000, OT_NONE, OT_NONE, 0x00, OI_CMPLTPS, OI_CMPLEPS },
	/*0x0096*/{ 0x30, OI_CMPEQPD, OT_XMM, OT_XMM128, 0x01052000, OT_NONE, OT_NONE, 0x00, OI_CMPLTPD, OI_CMPLEPD },
	/*0x0097*/{ 0x28, OI_CMPEQSS, OT_XMM, OT_XMM32, 0x01052000, OT_NONE, OT_NONE, 0x00, OI_CMPLTSS, OI_CMPLESS },
	/*0x0098*/{ 0x30, OI_CMPEQSD, OT_XMM, OT_XMM64, 0x01052000, OT_NONE, OT_NONE, 0x00, OI_CMPLTSD, OI_CMPLESD },
	/*0x0099*/{ 0x80, OI_VCMPEQPS, OT_YXMM, OT_VYXMM, 0x01093000, OT_YXMM128_256, OT_NONE, 0x01, OI_VCMPLTPS, OI_VCMPLEPS },
	/*0x009A*/{ 0x80, OI_VCMPEQPD, OT_YXMM, OT_VYXMM, 0x01093000, OT_YXMM128_256, OT_NONE, 0x01, OI_VCMPLTPD, OI_VCMPLEPD },
	/*0x009B*/{ 0x80, OI_VCMPEQSS, OT_XMM, OT_VXMM, 0x01093000, OT_XMM32, OT_NONE, 0x00, OI_VCMPLTSS, OI_VCMPLESS },
	/*0x009C*/{ 0x80, OI_VCMPEQSD, OT_XMM, OT_VXMM, 0x01093000, OT_XMM64, OT_NONE, 0x00, OI_VCMPLTSD, OI_VCMPLESD },
	/*0x009D*/{ 0x28, OI_PINSRW, OT_MMX, OT_R32_M16, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x009E*/{ 0x30, OI_PINSRW, OT_XMM, OT_R32_M16, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x009F*/{ 0x80, OI_VPINSRW, OT_XMM, OT_VXMM, 0x00193000, OT_R32_M16, OT_IMM8, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00A0*/{ 0x28, OI_PEXTRW, OT_REG32, OT_MMX_RM, 0x00492000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00A1*/{ 0x30, OI_PEXTRW, OT_REG32, OT_XMM_RM, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00A2*/{ 0x80, OI_VPEXTRW, OT_REG32_64, OT_XMM, 0x04093000, OT_IMM8, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00A3*/{ 0x28, OI_SHUFPS, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00A4*/{ 0x30, OI_SHUFPD, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00A5*/{ 0x80, OI_VSHUFPS, OT_YXMM, OT_VYXMM, 0x00193000, OT_YXMM128_256, OT_IMM8, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00A6*/{ 0x80, OI_VSHUFPD, OT_YXMM, OT_VYXMM, 0x00193000, OT_YXMM128_256, OT_IMM8, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00A7*/{ 0x08, OI_CMPXCHG8B, OT_MEM64_128, OT_NONE, 0xCC012801, OT_NONE, OT_NONE, 0x00, OI_UNKNOWN, OI_CMPXCHG16B },
	/*0x00A8*/{ 0x80, OI_VADDSUBPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00A9*/{ 0x80, OI_VADDSUBPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00AA*/{ 0x80, OI_VPSRLW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00AB*/{ 0x80, OI_VPSRLD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00AC*/{ 0x80, OI_VPSRLQ, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00AD*/{ 0x80, OI_VPADDQ, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00AE*/{ 0x80, OI_VPMULLW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00AF*/{ 0x80, OI_VMOVQ, OT_XMM64, OT_XMM, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00B0*/{ 0x80, OI_VPMOVMSKB, OT_REG32_64, OT_XMM, 0x04413000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00B1*/{ 0x80, OI_VPSUBUSB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00B2*/{ 0x80, OI_VPSUBUSW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00B3*/{ 0x80, OI_VPMINUB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00B4*/{ 0x80, OI_VPAND, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00B5*/{ 0x80, OI_VPADDUSW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00B6*/{ 0x80, OI_VPADDUSW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00B7*/{ 0x80, OI_VPMAXUB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00B8*/{ 0x80, OI_VPANDN, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00B9*/{ 0x80, OI_VPAVGB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00BA*/{ 0x80, OI_VPSRAW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00BB*/{ 0x80, OI_VPSRAD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00BC*/{ 0x80, OI_VPAVGW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00BD*/{ 0x80, OI_VPMULHUW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00BE*/{ 0x80, OI_VPMULHW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00BF*/{ 0x80, OI_VCVTTPD2DQ, OT_XMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00C0*/{ 0x80, OI_VCVTDQ2PD, OT_YXMM, OT_LXMM64_128, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00C1*/{ 0x80, OI_VCVTPD2DQ, OT_XMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00C2*/{ 0x80, OI_VMOVNTDQ, OT_LMEM128_256, OT_YXMM, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00C3*/{ 0x80, OI_VPSUBSB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00C4*/{ 0x80, OI_VPSUBSW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00C5*/{ 0x80, OI_VPMINSW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00C6*/{ 0x80, OI_VPOR, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00C7*/{ 0x80, OI_VPADDSB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00C8*/{ 0x80, OI_VPADDSW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00C9*/{ 0x80, OI_VPMAXSW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00CA*/{ 0x80, OI_VPXOR, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00CB*/{ 0x80, OI_VLDDQU, OT_YXMM, OT_LMEM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00CC*/{ 0x80, OI_VPSLLW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00CD*/{ 0x80, OI_VPSLLD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00CE*/{ 0x80, OI_VPSLLQ, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00CF*/{ 0x80, OI_VPMULUDQ, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00D0*/{ 0x80, OI_VPMADDWD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00D1*/{ 0x80, OI_VPSADBW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00D2*/{ 0x80, OI_VMASKMOVDQU, OT_XMM, OT_XMM_RM, 0x00413000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00D3*/{ 0x80, OI_VPSUBB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00D4*/{ 0x80, OI_VPSUBW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00D5*/{ 0x80, OI_VPSUBD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00D6*/{ 0x80, OI_VPSUBQ, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00D7*/{ 0x80, OI_VPADDB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00D8*/{ 0x80, OI_VPADDW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00D9*/{ 0x80, OI_VPADDD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00DA*/{ 0x80, OI_VPSHUFB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00DB*/{ 0x80, OI_VPHADDW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00DC*/{ 0x80, OI_VPHADDD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00DD*/{ 0x80, OI_VPHADDSW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00DE*/{ 0x80, OI_VPMADDUBSW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00DF*/{ 0x80, OI_VPHSUBW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00E0*/{ 0x80, OI_VPHSUBD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00E1*/{ 0x80, OI_VPHSUBSW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00E2*/{ 0x80, OI_VPSIGNB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00E3*/{ 0x80, OI_VPSIGNW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00E4*/{ 0x80, OI_VPSIGND, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00E5*/{ 0x80, OI_VPMULHRSW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00E6*/{ 0x80, OI_VPERMILPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00E7*/{ 0x80, OI_VPERMILPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00E8*/{ 0x80, OI_VTESTPS, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00E9*/{ 0x80, OI_VTESTPD, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00EA*/{ 0x48, OI_PBLENDVB, OT_XMM, OT_XMM128, 0x00092000, OT_NONE, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00EB*/{ 0x48, OI_BLENDVPS, OT_XMM, OT_XMM128, 0x00092000, OT_NONE, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00EC*/{ 0x48, OI_BLENDVPD, OT_XMM, OT_XMM128, 0x00092000, OT_NONE, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00ED*/{ 0x80, OI_VPTEST, OT_YXMM, OT_YXMM128_256, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00EE*/{ 0x80, OI_VBROADCASTSS, OT_YXMM, OT_MEM32, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00EF*/{ 0x80, OI_VBROADCASTSD, OT_YMM, OT_MEM64, 0x00013000, OT_NONE, OT_NONE, 0x50, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00F0*/{ 0x80, OI_VBROADCASTF128, OT_YMM, OT_MEM128, 0x00013000, OT_NONE, OT_NONE, 0x50, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00F1*/{ 0x80, OI_VPABSB, OT_XMM, OT_XMM128, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00F2*/{ 0x80, OI_VPABSW, OT_XMM, OT_XMM128, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00F3*/{ 0x80, OI_VPABSD, OT_XMM, OT_XMM128, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00F4*/{ 0x80, OI_VPMOVSXBW, OT_XMM, OT_XMM64, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00F5*/{ 0x80, OI_VPMOVSXBD, OT_XMM, OT_XMM32, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00F6*/{ 0x80, OI_VPMOVSXBQ, OT_XMM, OT_XMM16, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00F7*/{ 0x80, OI_VPMOVSXWD, OT_XMM, OT_XMM64, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00F8*/{ 0x80, OI_VPMOVSXWQ, OT_XMM, OT_XMM32, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00F9*/{ 0x80, OI_VPMOVSXDQ, OT_XMM, OT_XMM64, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00FA*/{ 0x80, OI_VPMULDQ, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00FB*/{ 0x80, OI_VPCMPEQQ, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00FC*/{ 0x80, OI_VMOVNTDQA, OT_XMM, OT_MEM128, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00FD*/{ 0x80, OI_VPACKUSDW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00FE*/{ 0x80, OI_VMASKMOVPS, OT_YXMM, OT_VYXMM, 0x00093000, OT_LMEM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x00FF*/{ 0x80, OI_VMASKMOVPD, OT_YXMM, OT_VYXMM, 0x00093000, OT_LMEM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0100*/{ 0x80, OI_VMASKMOVPS, OT_LMEM128_256, OT_VYXMM, 0x00093000, OT_YXMM, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0101*/{ 0x80, OI_VMASKMOVPD, OT_LMEM128_256, OT_VYXMM, 0x00093000, OT_YXMM, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0102*/{ 0x80, OI_VPMOVZXBW, OT_XMM, OT_XMM64, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0103*/{ 0x80, OI_VPMOVZXBD, OT_XMM, OT_XMM32, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0104*/{ 0x80, OI_VPMOVZXBQ, OT_XMM, OT_XMM16, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0105*/{ 0x80, OI_VPMOVZXWD, OT_XMM, OT_XMM64, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0106*/{ 0x80, OI_VPMOVZXWQ, OT_XMM, OT_XMM32, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0107*/{ 0x80, OI_VPMOVZXDQ, OT_XMM, OT_XMM64, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0108*/{ 0x80, OI_VPCMPGTQ, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0109*/{ 0x80, OI_VPMINSB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x010A*/{ 0x80, OI_VPMINSD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x010B*/{ 0x80, OI_VPMINUW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x010C*/{ 0x80, OI_VPMINUD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x010D*/{ 0x80, OI_VPMAXSB, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x010E*/{ 0x80, OI_VPMAXSD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x010F*/{ 0x80, OI_VPMAXUW, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0110*/{ 0x80, OI_VPMAXUD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0111*/{ 0x80, OI_VPMULLD, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0112*/{ 0x80, OI_VPHMINPOSUW, OT_XMM, OT_XMM128, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0113*/{ 0x88, OI_VFMADDSUB132PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFMADDSUB132PD, OI_UNKNOWN },
	/*0x0114*/{ 0x88, OI_VFMSUBADD132PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFMSUBADD132PD, OI_UNKNOWN },
	/*0x0115*/{ 0x88, OI_VFMADD132PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFMADD132PD, OI_UNKNOWN },
	/*0x0116*/{ 0x88, OI_VFMADD132SS, OT_XMM, OT_VXMM, 0x00093000, OT_WXMM32_64, OT_NONE, 0x06, OI_VFMADD132SD, OI_UNKNOWN },
	/*0x0117*/{ 0x88, OI_VFMSUB132PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFMSUB132PD, OI_UNKNOWN },
	/*0x0118*/{ 0x88, OI_VFMSUB132SS, OT_XMM, OT_VXMM, 0x00093000, OT_WXMM32_64, OT_NONE, 0x06, OI_VFMSUB132SD, OI_UNKNOWN },
	/*0x0119*/{ 0x88, OI_VFNMADD132PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFNMADD132PD, OI_UNKNOWN },
	/*0x011A*/{ 0x88, OI_VFNMADD132SS, OT_XMM, OT_VXMM, 0x00093000, OT_WXMM32_64, OT_NONE, 0x06, OI_VFNMADD132SD, OI_UNKNOWN },
	/*0x011B*/{ 0x88, OI_VFNMSUB132PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFNMSUB132PD, OI_UNKNOWN },
	/*0x011C*/{ 0x88, OI_VFNMSUB132SS, OT_XMM, OT_VXMM, 0x00093000, OT_WXMM32_64, OT_NONE, 0x06, OI_VFNMSUB132SD, OI_UNKNOWN },
	/*0x011D*/{ 0x88, OI_VFMADDSUB213PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFMADDSUB213PD, OI_UNKNOWN },
	/*0x011E*/{ 0x88, OI_VFMSUBADD213PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFMSUBADD213PD, OI_UNKNOWN },
	/*0x011F*/{ 0x88, OI_VFMADD213PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFMADD213PD, OI_UNKNOWN },
	/*0x0120*/{ 0x88, OI_VFMADD213SS, OT_XMM, OT_VXMM, 0x00093000, OT_WXMM32_64, OT_NONE, 0x06, OI_VFMADD213SD, OI_UNKNOWN },
	/*0x0121*/{ 0x88, OI_VFMSUB213PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFMSUB213PD, OI_UNKNOWN },
	/*0x0122*/{ 0x88, OI_VFMSUB213SS, OT_XMM, OT_VXMM, 0x00093000, OT_WXMM32_64, OT_NONE, 0x06, OI_VFMSUB213SD, OI_UNKNOWN },
	/*0x0123*/{ 0x88, OI_VFNMADD213PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFNMADD213PD, OI_UNKNOWN },
	/*0x0124*/{ 0x88, OI_VFNMADD213SS, OT_XMM, OT_VXMM, 0x00093000, OT_WXMM32_64, OT_NONE, 0x06, OI_VFNMADD213SD, OI_UNKNOWN },
	/*0x0125*/{ 0x88, OI_VFNMSUB213PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFNMSUB213PD, OI_UNKNOWN },
	/*0x0126*/{ 0x88, OI_VFNMSUB213SS, OT_XMM, OT_VXMM, 0x00093000, OT_WXMM32_64, OT_NONE, 0x06, OI_VFNMSUB213SD, OI_UNKNOWN },
	/*0x0127*/{ 0x88, OI_VFMADDSUB231PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFMADDSUB231PD, OI_UNKNOWN },
	/*0x0128*/{ 0x88, OI_VFMSUBADD231PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFMSUBADD231PD, OI_UNKNOWN },
	/*0x0129*/{ 0x88, OI_VFMADD231PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFMADD231PD, OI_UNKNOWN },
	/*0x012A*/{ 0x88, OI_VFMADD231SS, OT_XMM, OT_VXMM, 0x00093000, OT_WXMM32_64, OT_NONE, 0x06, OI_VFMADD231SD, OI_UNKNOWN },
	/*0x012B*/{ 0x88, OI_VFMSUB231PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFMSUB231PD, OI_UNKNOWN },
	/*0x012C*/{ 0x88, OI_VFMSUB231SS, OT_XMM, OT_VXMM, 0x00093000, OT_WXMM32_64, OT_NONE, 0x06, OI_VFMSUB231SD, OI_UNKNOWN },
	/*0x012D*/{ 0x88, OI_VFNMADD231PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFNMADD231PD, OI_UNKNOWN },
	/*0x012E*/{ 0x88, OI_VFNMADD231SS, OT_XMM, OT_VXMM, 0x00093000, OT_WXMM32_64, OT_NONE, 0x06, OI_VFNMADD231SD, OI_VFNMADD231SD },
	/*0x012F*/{ 0x88, OI_VFNMSUB231PS, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x07, OI_VFNMSUB231PD, OI_UNKNOWN },
	/*0x0130*/{ 0x88, OI_VFNMSUB231SS, OT_XMM, OT_VXMM, 0x00093000, OT_WXMM32_64, OT_NONE, 0x06, OI_VFNMSUB231SD, OI_UNKNOWN },
	/*0x0131*/{ 0x98, OI_VAESIMC, OT_XMM, OT_XMM128, 0x00013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0132*/{ 0x98, OI_VAESENC, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0133*/{ 0x98, OI_VAESENCLAST, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0134*/{ 0x98, OI_VAESDEC, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0135*/{ 0x98, OI_VAESDECLAST, OT_XMM, OT_VXMM, 0x00093000, OT_XMM128, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0136*/{ 0x80, OI_VPERMILPS, OT_YXMM, OT_YXMM128_256, 0x00093000, OT_IMM8, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0137*/{ 0x80, OI_VPERMILPD, OT_YXMM, OT_YXMM128_256, 0x00093000, OT_IMM8, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0138*/{ 0x80, OI_VPERM2F128, OT_YMM, OT_VYMM, 0x00193000, OT_YMM256, OT_IMM8, 0x10, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0139*/{ 0x48, OI_ROUNDPS, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x013A*/{ 0x80, OI_VROUNDPS, OT_YXMM, OT_YXMM128_256, 0x00093000, OT_IMM8, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x013B*/{ 0x48, OI_ROUNDPD, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x013C*/{ 0x80, OI_VROUNDPD, OT_YXMM, OT_YXMM128_256, 0x00093000, OT_IMM8, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x013D*/{ 0x48, OI_ROUNDSS, OT_XMM, OT_XMM32, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x013E*/{ 0x80, OI_VROUNDSS, OT_XMM, OT_VXMM, 0x00193000, OT_XMM32, OT_IMM8, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x013F*/{ 0x48, OI_ROUNDSD, OT_XMM, OT_XMM64, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0140*/{ 0x80, OI_VROUNDSD, OT_XMM, OT_VXMM, 0x00193000, OT_XMM64, OT_IMM8, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0141*/{ 0x48, OI_BLENDPS, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0142*/{ 0x80, OI_VBLENDPS, OT_YXMM, OT_VYXMM, 0x00193000, OT_YXMM128_256, OT_IMM8, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0143*/{ 0x48, OI_BLENDPD, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0144*/{ 0x80, OI_VBLENDPD, OT_YXMM, OT_VYXMM, 0x00193000, OT_YXMM128_256, OT_IMM8, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0145*/{ 0x48, OI_PBLENDW, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0146*/{ 0x80, OI_VPBLENDW, OT_XMM, OT_VXMM, 0x00193000, OT_XMM128, OT_IMM8, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0147*/{ 0x40, OI_PALIGNR, OT_MMX, OT_MMX64, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0148*/{ 0x40, OI_PALIGNR, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0149*/{ 0x80, OI_VPALIGNR, OT_XMM, OT_VXMM, 0x00193000, OT_XMM128, OT_IMM8, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x014A*/{ 0x48, OI_PEXTRB, OT_R32_64_M8, OT_XMM, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x014B*/{ 0x80, OI_VPEXTRB, OT_REG32_64_M8, OT_XMM, 0x04093000, OT_IMM8, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x014C*/{ 0x48, OI_PEXTRW, OT_R32_64_M16, OT_XMM, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x014D*/{ 0x80, OI_VPEXTRW, OT_REG32_64_M16, OT_XMM, 0x04093000, OT_IMM8, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x014E*/{ 0x48, OI_PEXTRD, OT_RM32_64, OT_XMM, 0x08092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_PEXTRQ },
	/*0x014F*/{ 0x80, OI_VPEXTRD, OT_WRM32_64, OT_XMM, 0x00093000, OT_IMM8, OT_NONE, 0x46, OI_VPEXTRQ, OI_UNKNOWN },
	/*0x0150*/{ 0x48, OI_EXTRACTPS, OT_RM32_64, OT_XMM, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0151*/{ 0x80, OI_VEXTRACTPS, OT_RM32, OT_XMM, 0x00093000, OT_IMM8, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0152*/{ 0x80, OI_VINSERTF128, OT_YMM, OT_VYMM, 0x00193000, OT_XMM128, OT_IMM8, 0x10, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0153*/{ 0x80, OI_VEXTRACTF128, OT_XMM128, OT_YMM, 0x00093000, OT_IMM8, OT_NONE, 0x50, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0154*/{ 0x48, OI_PINSRB, OT_XMM, OT_R32_M8, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0155*/{ 0x80, OI_VPINSRB, OT_XMM, OT_VXMM, 0x00193000, OT_REG32_64_M8, OT_IMM8, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0156*/{ 0x48, OI_INSERTPS, OT_XMM, OT_XMM32, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0157*/{ 0x80, OI_VINSERTPS, OT_XMM, OT_VXMM, 0x00193000, OT_XMM32, OT_IMM8, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0158*/{ 0x48, OI_PINSRD, OT_XMM, OT_RM32_64, 0x08092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_PINSRQ },
	/*0x0159*/{ 0x80, OI_VPINSRD, OT_XMM, OT_VXMM, 0x00193000, OT_WRM32_64, OT_IMM8, 0x06, OI_VPINSRQ, OI_UNKNOWN },
	/*0x015A*/{ 0x48, OI_DPPS, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x015B*/{ 0x80, OI_VDPPS, OT_YXMM, OT_VYXMM, 0x00193000, OT_YXMM128_256, OT_IMM8, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x015C*/{ 0x48, OI_DPPD, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x015D*/{ 0x80, OI_VDPPD, OT_XMM, OT_VXMM, 0x00193000, OT_XMM128, OT_IMM8, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x015E*/{ 0x48, OI_MPSADBW, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x015F*/{ 0x80, OI_VMPSADBW, OT_XMM, OT_VXMM, 0x00193000, OT_XMM128, OT_IMM8, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0160*/{ 0x90, OI_PCLMULQDQ, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0161*/{ 0x90, OI_VPCLMULQDQ, OT_XMM, OT_VXMM, 0x00193000, OT_XMM128, OT_IMM8, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0162*/{ 0x80, OI_VBLENDVPS, OT_YXMM, OT_VYXMM, 0x00193000, OT_YXMM128_256, OT_YXMM_IMM, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0163*/{ 0x80, OI_VBLENDVPD, OT_YXMM, OT_VYXMM, 0x00193000, OT_YXMM128_256, OT_YXMM_IMM, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0164*/{ 0x80, OI_VPBLENDVB, OT_XMM, OT_VXMM, 0x00193000, OT_XMM128, OT_XMM_IMM, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0165*/{ 0x50, OI_PCMPESTRM, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0166*/{ 0x80, OI_VPCMPESTRM, OT_XMM, OT_XMM128, 0x00093000, OT_IMM8, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0167*/{ 0x50, OI_PCMPESTRI, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0168*/{ 0x80, OI_VPCMPESTRI, OT_XMM, OT_XMM128, 0x00093000, OT_IMM8, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0169*/{ 0x50, OI_PCMPISTRM, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x016A*/{ 0x80, OI_VPCMPISTRM, OT_XMM, OT_XMM128, 0x00093000, OT_IMM8, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x016B*/{ 0x50, OI_PCMPISTRI, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x016C*/{ 0x80, OI_VPCMPISTRI, OT_XMM, OT_XMM128, 0x00093000, OT_IMM8, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x016D*/{ 0x98, OI_AESKEYGENASSIST, OT_XMM, OT_XMM128, 0x00092000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x016E*/{ 0x98, OI_VAESKEYGENASSIST, OT_XMM, OT_XMM128, 0x00093000, OT_IMM8, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x016F*/{ 0x80, OI_VPSRLW, OT_VXMM, OT_XMM_RM, 0x40493000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0170*/{ 0x80, OI_VPSRAW, OT_VXMM, OT_XMM_RM, 0x40493000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0171*/{ 0x80, OI_VPSLLW, OT_VXMM, OT_XMM_RM, 0x40493000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0172*/{ 0x80, OI_VPSRLD, OT_VXMM, OT_XMM_RM, 0x40493000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0173*/{ 0x80, OI_VPSRAD, OT_VXMM, OT_XMM_RM, 0x40493000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0174*/{ 0x80, OI_VPSLLD, OT_VXMM, OT_XMM_RM, 0x40493000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0175*/{ 0x80, OI_VPSRLQ, OT_VXMM, OT_XMM_RM, 0x40493000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0176*/{ 0x80, OI_VPSRLDQ, OT_VXMM, OT_XMM_RM, 0x40493000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0177*/{ 0x80, OI_VPSLLQ, OT_VXMM, OT_XMM_RM, 0x40493000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0178*/{ 0x80, OI_VPSLLDQ, OT_VXMM, OT_XMM_RM, 0x40493000, OT_IMM8, OT_NONE, 0x00, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0179*/{ 0x08, OI_FXSAVE, OT_MEM, OT_NONE, 0x4C012800, OT_NONE, OT_NONE, 0x00, OI_UNKNOWN, OI_FXSAVE64 },
	/*0x017A*/{ 0x08, OI_FXRSTOR, OT_MEM, OT_NONE, 0x4C012800, OT_NONE, OT_NONE, 0x00, OI_UNKNOWN, OI_FXRSTOR64 },
	/*0x017B*/{ 0x80, OI_VLDMXCSR, OT_MEM32, OT_NONE, 0x40013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x017C*/{ 0x80, OI_VSTMXCSR, OT_MEM32, OT_NONE, 0x40013000, OT_NONE, OT_NONE, 0x40, OI_UNKNOWN, OI_UNKNOWN },
	/*0x017D*/{ 0xA0, OI_VPERMQ, OT_YMM, OT_YXMM128_256, 0x00093000, OT_IMM8, OT_NONE, 0x50, OI_UNKNOWN, OI_UNKNOWN },
	/*0x017E*/{ 0xA0, OI_VPERMPD, OT_YMM, OT_YXMM128_256, 0x00093000, OT_IMM8, OT_NONE, 0x50, OI_UNKNOWN, OI_UNKNOWN },
	/*0x017F*/{ 0xA0, OI_VPBLENDD, OT_YXMM, OT_VYXMM, 0x00193000, OT_YXMM128_256, OT_IMM8, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0180*/{ 0xA0, OI_VPERM2I128, OT_YMM, OT_VYMM, 0x00193000, OT_YMM256, OT_IMM8, 0x10, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0181*/{ 0x80, OI_VCVTPS2PH, OT_LXMM64_128, OT_YXMM, 0x00093000, OT_IMM8, OT_NONE, 0x50, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0182*/{ 0x80, OI_VCVTPH2PS, OT_YXMM, OT_LXMM64_128, 0x00013000, OT_NONE, OT_NONE, 0x50, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0183*/{ 0xA0, OI_VINSERTI128, OT_YMM, OT_VYMM, 0x00193000, OT_LXMM64_128, OT_IMM8, 0x10, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0184*/{ 0xA0, OI_VEXTRACTI128, OT_LXMM64_128, OT_YMM, 0x00093000, OT_IMM8, OT_NONE, 0x50, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0185*/{ 0xA0, OI_VPERMPS, OT_YMM, OT_VYMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x10, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0186*/{ 0xA0, OI_VPERMD, OT_YMM, OT_VYMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x50, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0187*/{ 0xA0, OI_VPSRLVD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x40, OI_VPSRLVQ, OI_VPSRLVQ },
	/*0x0188*/{ 0xA0, OI_VPSRAVD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0189*/{ 0xA0, OI_VPSLLVD, OT_YXMM, OT_VYXMM, 0x00093000, OT_YXMM128_256, OT_NONE, 0x01, OI_VPSLLVQ, OI_VPSLLVQ },
	/*0x018A*/{ 0xA0, OI_VPBROADCASTD, OT_YXMM, OT_WXMM32_64, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x018B*/{ 0xA0, OI_VPBROADCASTQ, OT_YXMM, OT_WXMM32_64, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x018C*/{ 0xA0, OI_VBROADCASTI128, OT_YMM, OT_MEM128, 0x00013000, OT_NONE, OT_NONE, 0x50, OI_UNKNOWN, OI_UNKNOWN },
	/*0x018D*/{ 0xA0, OI_VPBROADCASTB, OT_YXMM, OT_WXMM32_64, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x018E*/{ 0xA0, OI_VPBROADCASTW, OT_YXMM, OT_WXMM32_64, 0x00013000, OT_NONE, OT_NONE, 0x41, OI_UNKNOWN, OI_UNKNOWN },
	/*0x018F*/{ 0xA0, OI_VPGATHERDD, OT_YXMM, OT_VSIB_M32_64, 0x00093000, OT_VYXMM, OT_NONE, 0x01, OI_VPGATHERDQ, OI_VPGATHERDQ },
	/*0x0190*/{ 0xA0, OI_VPGATHERQD, OT_YXMM, OT_VSIB_M32_64, 0x00093000, OT_VYXMM, OT_NONE, 0x01, OI_VPGATHERQQ, OI_VPGATHERQQ },
	/*0x0191*/{ 0xA0, OI_VGATHERDPS, OT_YXMM, OT_VSIB_M32_64, 0x00093000, OT_VYXMM, OT_NONE, 0x01, OI_VGATHERDPD, OI_VGATHERDPD },
	/*0x0192*/{ 0xA0, OI_VGATHERQPS, OT_YXMM, OT_VSIB_M32_64, 0x00093000, OT_VYXMM, OT_NONE, 0x01, OI_VGATHERQPD, OI_VGATHERQPD },
	/*0x0193*/{ 0xA0, OI_VPMASKMOVD, OT_YXMM, OT_VYXMM, 0x00093000, OT_LMEM128_256, OT_NONE, 0x40, OI_VPMASKMOVQ, OI_VPMASKMOVQ },
	/*0x0194*/{ 0xA0, OI_VPMASKMOVD, OT_YXMM, OT_VYXMM, 0x00093000, OT_LMEM128_256, OT_NONE, 0x40, OI_VPMASKMOVQ, OI_VPMASKMOVQ },
	/*0x0195*/{ 0xA1, OI_RORX, OT_REG32_64, OT_RM32_64, 0x00093000, OT_IMM8, OT_NONE, 0x10, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0196*/{ 0xA1, OI_ANDN, OT_REG32_64, OT_WREG32_64, 0x80093000, OT_RM32_64, OT_NONE, 0x02, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0197*/{ 0xA1, OI_BZHI, OT_REG32_64, OT_RM32_64, 0x80093000, OT_WREG32_64, OT_NONE, 0x02, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0198*/{ 0xA1, OI_BEXTR, OT_REG32_64, OT_RM32_64, 0x80093000, OT_WREG32_64, OT_NONE, 0x02, OI_UNKNOWN, OI_UNKNOWN },
	/*0x0199*/{ 0xA1, OI_PEXT, OT_REG32_64, OT_WREG32_64, 0x80093000, OT_RM32_64, OT_NONE, 0x02, OI_UNKNOWN, OI_UNKNOWN },
	/*0x019A*/{ 0xA1, OI_PDEP, OT_REG32_64, OT_WREG32_64, 0x80093000, OT_RM32_64, OT_NONE, 0x02, OI_UNKNOWN, OI_UNKNOWN },
	/*0x019B*/{ 0xA1, OI_MULX, OT_REG32_64, OT_WREG32_64, 0x80093000, OT_RM32_64, OT_NONE, 0x02, OI_UNKNOWN, OI_UNKNOWN },
	/*0x019C*/{ 0xA1, OI_SHLX, OT_REG32_64, OT_RM32_64, 0x80093000, OT_WREG32_64, OT_NONE, 0x02, OI_UNKNOWN, OI_UNKNOWN },
	/*0x019D*/{ 0xA1, OI_SARX, OT_REG32_64, OT_RM32_64, 0x80093000, OT_WREG32_64, OT_NONE, 0x02, OI_UNKNOWN, OI_UNKNOWN },
	/*0x019E*/{ 0xA1, OI_SHRX, OT_REG32_64, OT_RM32_64, 0x80093000, OT_WREG32_64, OT_NONE, 0x02, OI_UNKNOWN, OI_UNKNOWN },
	/*0x019F*/{ 0xA2, OI_BLSR, OT_WREG32_64, OT_RM32_64, 0xC0013000, OT_NONE, OT_NONE, 0x02, OI_UNKNOWN, OI_UNKNOWN },
	/*0x01A0*/{ 0xA2, OI_BLSMSK, OT_WREG32_64, OT_RM32_64, 0xC0013000, OT_NONE, OT_NONE, 0x02, OI_UNKNOWN, OI_UNKNOWN },
	/*0x01A1*/{ 0xA2, OI_BLSI, OT_WREG32_64, OT_RM32_64, 0xC0013000, OT_NONE, OT_NONE, 0x02, OI_UNKNOWN, OI_UNKNOWN },
};
